from enum import IntEnum

from .NamedObject import NamedObject


class Shader(NamedObject):
    def __init__(self, reader):
        super().__init__(reader=reader)
        version = reader.version
        if version[0] == 5 and version[1] >= 5 or version[0] > 5:  # 5.5 and up
            self.m_ParsedForm = SerializedShader(reader)
            self.platforms = [
                ShaderCompilerPlatform(x) for x in reader.read_u_int_array()
            ]

            # 2019.3 and up
            if version[0] > 2019 or (version[0] == 2019 and version[0] >= 3):
                offsets = reader.read_u_int_array_array()[0]
                compressedLengths = reader.read_u_int_array_array()[0]
                decompressedLengths = reader.read_u_int_array_array()[0]
            else:
                self.offsets = reader.read_u_int_array()
                self.compressedLengths = reader.read_u_int_array()
                self.decompressedLengths = reader.read_u_int_array()
            self.compressedBlob = reader.read_bytes(reader.read_int())
        else:
            self.m_Script = reader.read_bytes(reader.read_int())
            reader.align_stream()
            self.m_PathName = reader.read_aligned_string()
            if version[0] == 5 and version[1] >= 3:  # 5.3 - 5.4
                self.decompressedSize = reader.read_u_int()
                self.m_SubProgramBlob = reader.read_bytes(reader.read_int())


class StructParameter:
    def __init__(self, reader):
        self.m_NameIndex = reader.read_int()
        self.m_Index = reader.read_int()
        self.m_ArraySize = reader.read_int()
        self.m_StructSize = reader.read_int()

        numVectorParams = reader.read_int()
        self.m_VectorParams = [VectorParameter(reader) for _ in range(numVectorParams)]

        numMatrixParams = reader.read_int()
        self.m_MatrixParams = [MatrixParameter(reader) for _ in range(numMatrixParams)]


class SamplerParameter:
    def __init__(self, reader):
        self.sampler = reader.read_u_int()
        self.bindPoint = reader.read_int()


class TextureDimension(IntEnum):
    kTexDimUnknown = (-1,)
    kTexDimNone = (0,)
    kTexDimAny = (1,)
    kTexDim2D = (2,)
    kTexDim3D = (3,)
    kTexDimCUBE = (4,)
    kTexDim2DArray = (5,)
    kTexDimCubeArray = (6,)
    kTexDimForce32Bit = (2147483647,)


class SerializedTextureProperty:
    def __init__(self, reader):
        self.m_DefaultName = reader.read_aligned_string()
        self.m_TexDim = TextureDimension(reader.read_int())


class SerializedPropertyType(IntEnum):
    kColor = (0,)
    kVector = (1,)
    kFloat = (2,)
    kRange = (3,)
    kTexture = (4,)


class SerializedProperty:
    def __init__(self, reader):
        self.m_Name = reader.read_aligned_string()
        self.m_Description = reader.read_aligned_string()
        self.m_Attributes = reader.read_string_array()
        self.m_Type = SerializedPropertyType(reader.read_int())
        self.m_Flags = reader.read_u_int()
        self.m_DefValue = reader.read_float_array(4)
        self.m_DefTexture = SerializedTextureProperty(reader)


class SerializedProperties:
    def __init__(self, reader):
        numProps = reader.read_int()
        self.m_Props = [SerializedProperty(reader) for _ in range(numProps)]


class SerializedShaderFloatValue:
    def __init__(self, reader):
        self.val = reader.read_float()
        self.name = reader.read_aligned_string()


class SerializedShaderRTBlendState:
    def __init__(self, reader):
        self.srcBlend = SerializedShaderFloatValue(reader)
        self.destBlend = SerializedShaderFloatValue(reader)
        self.srcBlendAlpha = SerializedShaderFloatValue(reader)
        self.destBlendAlpha = SerializedShaderFloatValue(reader)
        self.blendOp = SerializedShaderFloatValue(reader)
        self.blendOpAlpha = SerializedShaderFloatValue(reader)
        self.colMask = SerializedShaderFloatValue(reader)


class SerializedStencilOp:
    def __init__(self, reader):
        self.pass_ = SerializedShaderFloatValue(reader)
        self.fail = SerializedShaderFloatValue(reader)
        self.zFail = SerializedShaderFloatValue(reader)
        self.comp = SerializedShaderFloatValue(reader)


class SerializedShaderVectorValue:
    def __init__(self, reader):
        self.x = SerializedShaderFloatValue(reader)
        self.y = SerializedShaderFloatValue(reader)
        self.z = SerializedShaderFloatValue(reader)
        self.w = SerializedShaderFloatValue(reader)
        self.name = reader.read_aligned_string()


class FogMode(IntEnum):
    kFogUnknown = (-1,)
    kFogDisabled = (0,)
    kFogLinear = (1,)
    kFogExp = (2,)
    kFogExp2 = (3,)


class SerializedShaderState:
    def __init__(self, reader):
        version = reader.version

        self.m_Name = reader.read_aligned_string()
        self.rtBlend = [SerializedShaderRTBlendState(reader) for _ in range(8)]
        self.rtSeparateBlend = reader.read_boolean()
        reader.align_stream()
        if version[0] > 2017 or (
            version[0] == 2017 and version[1] >= 2
        ):  # 2017.2 and up
            self.zClip = SerializedShaderFloatValue(reader)
        self.zTest = SerializedShaderFloatValue(reader)
        self.zWrite = SerializedShaderFloatValue(reader)
        self.culling = SerializedShaderFloatValue(reader)
        self.offsetFactor = SerializedShaderFloatValue(reader)
        self.offsetUnits = SerializedShaderFloatValue(reader)
        self.alphaToMask = SerializedShaderFloatValue(reader)
        self.stencilOp = SerializedStencilOp(reader)
        self.stencilOpFront = SerializedStencilOp(reader)
        self.stencilOpBack = SerializedStencilOp(reader)
        self.stencilReadMask = SerializedShaderFloatValue(reader)
        self.stencilWriteMask = SerializedShaderFloatValue(reader)
        self.stencilRef = SerializedShaderFloatValue(reader)
        self.fogStart = SerializedShaderFloatValue(reader)
        self.fogEnd = SerializedShaderFloatValue(reader)
        self.fogDensity = SerializedShaderFloatValue(reader)
        self.fogColor = SerializedShaderVectorValue(reader)
        self.fogMode = FogMode(reader.read_int())
        self.gpuProgramID = reader.read_int()
        self.m_Tags = SerializedTagMap(reader)
        self.m_LOD = reader.read_int()
        self.lighting = reader.read_boolean()
        reader.align_stream()


class ShaderBindChannel:
    def __init__(self, reader):
        self.source = reader.read_byte()
        self.target = reader.read_byte()


class ParserBindChannels:
    def __init__(self, reader):
        numChannels = reader.read_int()
        self.m_Channels = [ShaderBindChannel(reader) for _ in range(numChannels)]
        reader.align_stream()
        self.m_SourceMap = reader.read_u_int()


class VectorParameter:
    def __init__(self, reader):
        self.m_NameIndex = reader.read_int()
        self.m_Index = reader.read_int()
        self.m_ArraySize = reader.read_int()
        self.m_Type = reader.read_byte()
        self.m_Dim = reader.read_byte()
        reader.align_stream()


class MatrixParameter:
    def __init__(self, reader):
        self.m_NameIndex = reader.read_int()
        self.m_Index = reader.read_int()
        self.m_ArraySize = reader.read_int()
        self.m_Type = reader.read_byte()
        self.m_RowCount = reader.read_byte()
        reader.align_stream()


class TextureParameter:
    def __init__(self, reader):
        version = reader.version
        self.m_NameIndex = reader.read_int()
        self.m_Index = reader.read_int()
        self.m_SamplerIndex = reader.read_int()
        if version[0] > 2017 or (
            version[0] == 2017 and version[1] >= 3
        ):  # 2017.3 and up
            self.m_MultiSampled = reader.read_boolean()
        self.m_Dim = reader.read_byte()
        reader.align_stream()


class BufferBinding:
    def __init__(self, reader):
        self.m_NameIndex = reader.read_int()
        self.m_Index = reader.read_int()


class ConstantBuffer:
    def __init__(self, reader):
        version = reader.version

        self.m_NameIndex = reader.read_int()

        numMatrixParams = reader.read_int()
        self.m_MatrixParams = [MatrixParameter(reader) for _ in range(numMatrixParams)]

        numVectorParams = reader.read_int()
        self.m_VectorParams = [VectorParameter(reader) for _ in range(numVectorParams)]
        if version[0] > 2017 or (
            version[0] == 2017 and version[1] >= 3
        ):  # 2017.3 and up
            numStructParams = reader.read_int()
            self.m_StructParams = [
                StructParameter(reader) for _ in range(numStructParams)
            ]
        self.m_Size = reader.read_int()


class UAVParameter:
    def __init__(self, reader):
        self.m_NameIndex = reader.read_int()
        self.m_Index = reader.read_int()
        self.m_OriginalIndex = reader.read_int()


class ShaderGpuProgramType(IntEnum):
    kShaderGpuProgramUnknown = (0,)
    kShaderGpuProgramGLLegacy = (1,)
    kShaderGpuProgramGLES31AEP = (2,)
    kShaderGpuProgramGLES31 = (3,)
    kShaderGpuProgramGLES3 = (4,)
    kShaderGpuProgramGLES = (5,)
    kShaderGpuProgramGLCore32 = (6,)
    kShaderGpuProgramGLCore41 = (7,)
    kShaderGpuProgramGLCore43 = (8,)
    kShaderGpuProgramDX9VertexSM20 = (9,)
    kShaderGpuProgramDX9VertexSM30 = (10,)
    kShaderGpuProgramDX9PixelSM20 = (11,)
    kShaderGpuProgramDX9PixelSM30 = (12,)
    kShaderGpuProgramDX10Level9Vertex = (13,)
    kShaderGpuProgramDX10Level9Pixel = (14,)
    kShaderGpuProgramDX11VertexSM40 = (15,)
    kShaderGpuProgramDX11VertexSM50 = (16,)
    kShaderGpuProgramDX11PixelSM40 = (17,)
    kShaderGpuProgramDX11PixelSM50 = (18,)
    kShaderGpuProgramDX11GeometrySM40 = (19,)
    kShaderGpuProgramDX11GeometrySM50 = (20,)
    kShaderGpuProgramDX11HullSM50 = (21,)
    kShaderGpuProgramDX11DomainSM50 = (22,)
    kShaderGpuProgramMetalVS = (23,)
    kShaderGpuProgramMetalFS = (24,)
    kShaderGpuProgramSPIRV = (25,)
    kShaderGpuProgramConsole = (26,)


class SerializedSubProgram:
    def __init__(self, reader):
        version = reader.version

        self.m_BlobIndex = reader.read_u_int()
        self.m_Channels = ParserBindChannels(reader)

        if version[0] >= 2019:  # 2019 and up
            self.m_GlobalKeywordIndices = reader.read_u_short_array()
            reader.align_stream()
            self.m_LocalKeywordIndices = reader.read_u_short_array()
            reader.align_stream()
        else:
            self.m_KeywordIndices = reader.read_u_short_array()
            if version[0] >= 2017:  # 2017 and up
                reader.align_stream()

        self.m_ShaderHardwareTier = reader.read_byte()
        self.m_GpuProgramType = ShaderGpuProgramType(reader.read_byte())
        reader.align_stream()

        numVectorParams = reader.read_int()
        self.m_VectorParams = [VectorParameter(reader) for _ in range(numVectorParams)]

        numMatrixParams = reader.read_int()
        self.m_MatrixParams = [MatrixParameter(reader) for _ in range(numMatrixParams)]

        numTextureParams = reader.read_int()
        self.m_TextureParams = [
            TextureParameter(reader) for _ in range(numTextureParams)
        ]

        numBufferParams = reader.read_int()
        self.m_BufferParams = [BufferBinding(reader) for _ in range(numBufferParams)]

        numConstantBuffers = reader.read_int()
        self.m_ConstantBuffers = [
            ConstantBuffer(reader) for _ in range(numConstantBuffers)
        ]

        numConstantBufferBindings = reader.read_int()
        self.m_ConstantBufferBindings = [
            BufferBinding(reader) for _ in range(numConstantBufferBindings)
        ]

        numUAVParams = reader.read_int()
        self.m_UAVParams = [UAVParameter(reader) for _ in range(numUAVParams)]

        if version[0] >= 2017:  # 2017 and up
            numSamplers = reader.read_int()
            self.m_Samplers = [SamplerParameter(reader) for _ in range(numSamplers)]
        if version[0] > 2017 or (
            version[0] == 2017 and version[1] >= 2
        ):  # 2017.2 and up
            self.m_ShaderRequirements = reader.read_int()


class SerializedProgram:
    def __init__(self, reader):
        numSubPrograms = reader.read_int()
        self.m_SubPrograms = [
            SerializedSubProgram(reader) for _ in range(numSubPrograms)
        ]


class PassType(IntEnum):
    kPassTypeNormal = (0,)
    kPassTypeUse = (1,)
    kPassTypeGrab = 2


class SerializedPass:
    def __init__(self, reader):
        version = reader.version
        numIndices = reader.read_int()
        self.m_NameIndices = {}
        for _ in range(numIndices):
            key = reader.read_aligned_string()
            self.m_NameIndices[key] = reader.read_int()
        self.m_Type = PassType(reader.read_int())
        self.m_State = SerializedShaderState(reader)
        self.m_ProgramMask = reader.read_u_int()
        self.progVertex = SerializedProgram(reader)
        self.progFragment = SerializedProgram(reader)
        self.progGeometry = SerializedProgram(reader)
        self.progHull = SerializedProgram(reader)
        self.progDomain = SerializedProgram(reader)
        if version[0] > 2019 or (
            version[0] == 2019 and version[0] >= 3
        ):  # 2019.3 and up
            self.progRayTracing = SerializedProgram(reader)
        self.m_HasInstancingVariant = reader.read_boolean()
        if version[0] >= 2018:  # 2018 and up
            self.m_HasProceduralInstancingVariant = reader.read_boolean()
        reader.align_stream()
        self.m_UseName = reader.read_aligned_string()
        self.m_Name = reader.read_aligned_string()
        self.m_TextureName = reader.read_aligned_string()
        self.m_Tags = SerializedTagMap(reader)


class SerializedTagMap:
    def __init__(self, reader):
        numTags = reader.read_int()
        self.tags = {}
        for _ in range(numTags):
            key = reader.read_aligned_string()
            self.tags[key] = reader.read_aligned_string()


class SerializedSubShader:
    def __init__(self, reader):
        numPasses = reader.read_int()
        self.m_Passes = [SerializedPass(reader) for _ in range(numPasses)]
        self.m_Tags = SerializedTagMap(reader)
        self.m_LOD = reader.read_int()


class SerializedShaderDependency:
    def __init__(self, reader):
        self.from_ = reader.read_aligned_string()
        self.to = reader.read_aligned_string()


class SerializedShader:
    def __init__(self, reader):
        self.m_PropInfo = SerializedProperties(reader)
        numSubShaders = reader.read_int()
        self.m_SubShaders = [SerializedSubShader(reader) for _ in range(numSubShaders)]
        self.m_Name = reader.read_aligned_string()
        self.m_CustomEditorName = reader.read_aligned_string()
        self.m_FallbackName = reader.read_aligned_string()
        numDependencies = reader.read_int()
        self.m_Dependencies = [
            SerializedShaderDependency(reader) for _ in range(numDependencies)
        ]
        self.m_DisableNoSubshadersMessage = reader.read_boolean()
        reader.align_stream()


class ShaderCompilerPlatform(IntEnum):
    kShaderCompPlatformNone = (-1,)
    kShaderCompPlatformGL = (0,)
    kShaderCompPlatformD3D9 = (1,)
    kShaderCompPlatformXbox360 = (2,)
    kShaderCompPlatformPS3 = (3,)
    kShaderCompPlatformD3D11 = (4,)
    kShaderCompPlatformGLES20 = (5,)
    kShaderCompPlatformNaCl = (6,)
    kShaderCompPlatformFlash = (7,)
    kShaderCompPlatformD3D11_9x = (8,)
    kShaderCompPlatformGLES3Plus = (9,)
    kShaderCompPlatformPSP2 = (10,)
    kShaderCompPlatformPS4 = (11,)
    kShaderCompPlatformXboxOne = (12,)
    kShaderCompPlatformPSM = (13,)
    kShaderCompPlatformMetal = (14,)
    kShaderCompPlatformOpenGLCore = (15,)
    kShaderCompPlatformN3DS = (16,)
    kShaderCompPlatformWiiU = (17,)
    kShaderCompPlatformVulkan = (18,)
    kShaderCompPlatformSwitch = (19,)
    kShaderCompPlatformXboxOneD3D12 = (20,)



from enum import IntEnum

from PIL import Image, ImageDraw

from .Texture2DConverter import get_image_from_texture2d
from ..streams import EndianBinaryReader


# should be imported from Sprite, but too lazy to fix the import issues caused by that
class SpritePackingRotation(IntEnum):
    kSPRNone = (0,)
    kSPRFlipHorizontal = (1,)
    kSPRFlipVertical = (2,)
    kSPRRotate180 = (3,)
    kSPRRotate90 = 4


class SpritePackingMode(IntEnum):
    kSPMTight = (0,)
    kSPMRectangle = 1


def get_image(sprite, texture, alpha_texture) -> Image:
    if alpha_texture and getattr(alpha_texture, "type", "") == "Texture2D":
        cache_id = (texture.path_id, alpha_texture.path_id)
        if cache_id not in sprite.assets_file._cache:
            original_image = get_image_from_texture2d(texture.read(), False)
            alpha_image = get_image_from_texture2d(alpha_texture.read(), False)
            original_image = Image.merge(
                "RGBA", (*original_image.split()[:3], alpha_image.split()[0])
            )
            sprite.assets_file._cache[cache_id] = original_image
    else:
        cache_id = texture.path_id
        if cache_id not in sprite.assets_file._cache:
            original_image = get_image_from_texture2d(texture.read(), False)
            sprite.assets_file._cache[cache_id] = original_image
    return sprite.assets_file._cache[cache_id]


def get_image_from_sprite(m_Sprite) -> Image:
    atlas = None
    if m_Sprite.m_SpriteAtlas:
        atlas = m_Sprite.m_SpriteAtlas.read()
    elif m_Sprite.m_AtlasTags:
        # looks like the direct pointer is empty, let's try to find the Atlas via its name
        for obj in m_Sprite.assets_file.objects.values():
            if obj.type == "SpriteAtlas":
                atlas = obj.read()
                if atlas.name == m_Sprite.m_AtlasTags[0]:
                    break
                atlas = None

    if atlas:
        sprite_atlas_data = atlas.render_data_map[m_Sprite.m_RenderDataKey]
    else:
        sprite_atlas_data = m_Sprite.m_RD

    m_Texture2D = sprite_atlas_data.texture
    alpha_texture = sprite_atlas_data.alphaTexture
    texture_rect = sprite_atlas_data.textureRect
    texture_rect_offset = sprite_atlas_data.textureRectOffset
    settings_raw = sprite_atlas_data.settingsRaw

    original_image = get_image(m_Sprite, m_Texture2D, alpha_texture)

    sprite_image = original_image.crop(
        (texture_rect.left, texture_rect.top, texture_rect.right, texture_rect.bottom)
    )

    if settings_raw.packed == 1:
        rotation = settings_raw.packingRotation
        if rotation == SpritePackingRotation.kSPRFlipHorizontal:
            sprite_image = sprite_image.transpose(Image.FLIP_TOP_BOTTOM)
        # spriteImage = RotateFlip(RotateFlipType.RotateNoneFlipX)
        elif rotation == SpritePackingRotation.kSPRFlipVertical:
            sprite_image = sprite_image.transpose(Image.FLIP_LEFT_RIGHT)
        # spriteImage.RotateFlip(RotateFlipType.RotateNoneFlipY)
        elif rotation == SpritePackingRotation.kSPRRotate180:
            sprite_image = sprite_image.transpose(Image.ROTATE_180)
        # spriteImage.RotateFlip(RotateFlipType.Rotate180FlipNone)
        elif rotation == SpritePackingRotation.kSPRRotate90:
            sprite_image = sprite_image.transpose(Image.ROTATE_270)
    # spriteImage.RotateFlip(RotateFlipType.Rotate270FlipNone)

    if settings_raw.packingMode == SpritePackingMode.kSPMTight:
        # Tight

        # create mask to keep only the polygon
        mask = Image.new("1", sprite_image.size, color=0)
        draw = ImageDraw.ImageDraw(mask)
        for triangle in get_triangles(m_Sprite):
            draw.polygon(triangle, fill=1)

        # apply the mask
        if sprite_image.mode == "RGBA":
            # the image already has an alpha channel,
            # so we have to use composite to keep it
            empty_img = Image.new(sprite_image.mode, sprite_image.size, color=0)
            sprite_image = Image.composite(sprite_image, empty_img, mask)
        else:
            # add mask as alpha-channel to keep the polygon clean
            sprite_image.putalpha(mask)

    return sprite_image.transpose(Image.FLIP_TOP_BOTTOM)


def get_triangles(m_Sprite):
    """
    returns the triangles of the sprite polygon
    """
    m_RD = m_Sprite.m_RD

    # read the raw points
    points = []
    if hasattr(m_RD, "vertices"):  # 5.6 down
        vertices = [v.pos for v in m_RD.vertices]
        points = [vertices[index] for index in range(m_RD.indices)]
    else:  # 5.6 and up
        m_Channel = m_RD.m_VertexData.m_Channels[0]  # kShaderChannelVertex
        m_Stream = m_RD.m_VertexData.m_Streams[m_Channel.stream]

        vertexReader = EndianBinaryReader(m_RD.m_VertexData.m_DataSize, endian="<")
        indexReader = EndianBinaryReader(m_RD.m_IndexBuffer, endian="<")

        for subMesh in m_RD.m_SubMeshes:
            vertexReader.Position = (
                m_Stream.offset
                + subMesh.firstVertex * m_Stream.stride
                + m_Channel.offset
            )

            vertices = []
            for _ in range(subMesh.vertexCount):
                vertices.append(vertexReader.read_vector3())
                vertexReader.Position = vertexReader.Position + m_Stream.stride - 12

            indexReader.Position = subMesh.firstByte

            for _ in range(subMesh.indexCount):
                points.append(
                    vertices[indexReader.read_u_short() - subMesh.firstVertex]
                )

    # normalize the points
    #  shift the whole point matrix into the positive space
    #  multiply them with a factor to scale them to the image
    min_x = min(p.X for p in points)
    min_y = min(p.Y for p in points)
    factor = m_Sprite.m_PixelsToUnits
    points = [((p.X - min_x) * factor, (p.Y - min_y) * factor) for p in points]

    # generate triangles from the given points
    return [points[i : i + 3] for i in range(0, len(points), 3)]



class Progress:
    pre_value: int

    def reset(self):
        self.pre_value = 0
        self.report(0)

    def report(self, current: int, total: int = None):
        if total:
            value = current * 100 // total
        else:
            value = current

        if value > self.pre_value:
            self.pre_value = value
            print(f"{value}%")



from .Component import Component
from .PPtr import PPtr


class Transform(Component):
    def __init__(self, reader):
        super().__init__(reader=reader)
        self.local_rotation = reader.read_quaternion()
        self.local_position = reader.read_vector3()
        self.local_scale = reader.read_vector3()

        children_count = reader.read_int()
        self.children = [PPtr(reader) for _ in range(children_count)]
        self.father = PPtr(reader)



from .NamedObject import NamedObject
from .PPtr import PPtr
from .Sprite import SpriteSettings


class SpriteAtlas(NamedObject):
    def __init__(self, reader):
        super().__init__(reader=reader)
        packed_sprites_size = reader.read_int()
        self.PackSprites = [PPtr(reader) for _ in range(packed_sprites_size)]

        self.packed_sprite_names_to_index = reader.read_string_array()
        m_render_data_map_size = reader.read_int()
        self.render_data_map = {}
        for _ in range(m_render_data_map_size):
            first = reader.read_bytes(16)  # GUID
            second = reader.read_long()
            value = SpriteAtlasData(reader)
            self.render_data_map[(first, second)] = value


class SpriteAtlasData:
    def __init__(self, reader):
        self.version = reader.version
        self.texture = PPtr(reader)  # Texture2D
        self.alphaTexture = PPtr(reader)  # Texture2D
        self.textureRect = reader.read_rectangle_f()
        self.textureRectOffset = reader.read_vector2()
        if self.version[0] > 2017 or (
            self.version[0] == 2017 and self.version[1] >= 2
        ):  # 2017.2 and up
            self.atlasRectOffset = reader.read_vector2()
        self.uvTransform = reader.read_vector4()
        self.downscaleMultiplier = reader.read_float()
        self.settingsRaw = SpriteSettings(reader)



from sys import platform

import texture2ddecoder
from PIL import Image

from ..enums import TextureFormat, BuildTarget

TF = TextureFormat
PREFER_BASISU = False


def get_image_from_texture2d(texture_2d, flip=True) -> Image:
    """converts the given texture into PIL.Image

    :param texture_2d: texture to be converterd
    :type texture_2d: Texture2D
    :param flip: flips the image back to the original (all Unity textures are flipped by default)
    :type flip: bool
    :return: PIL.Image object
    :rtype: Image
    """
    image_data = texture_2d.image_data
    texture_format = texture_2d.m_TextureFormat

    selection = CONV_TABLE[texture_format]

    if len(selection) == 0:
        raise NotImplementedError(
            f"Not implemented texture format: {texture_format.name}"
        )

    if texture_format in XBOX_SWAP_FORMATS:
        image_data = swap_bytes_for_xbox(image_data, texture_2d.platform)

    if "Crunched" in texture_format.name:
        version = texture_2d.version
        if (version[0] > 2017 or (version[0] == 2017 and version[1] >= 3) #2017.3 and up
            or texture_format == TextureFormat.ETC_RGB4Crunched
            or texture_format == TextureFormat.ETC2_RGBA8Crunched):
            image_data = texture2ddecoder.unpack_unity_crunch(image_data)
        else:
            image_data = texture2ddecoder.unpack_crunch(image_data)

    img = selection[0](
        image_data, texture_2d.m_Width, texture_2d.m_Height, *selection[1:]
    )

    if img and flip:
        return img.transpose(Image.FLIP_TOP_BOTTOM)
    return img


def swap_bytes_for_xbox(image_data: bytes, build_target: BuildTarget) -> bytes:
    """swaps the texture bytes
    This is required for textures deployed on XBOX360.

    :param image_data: texture data
    :type image_data: bytes
    :param build_target: platform of the asset
    :type build_target: BuildTarget
    :return: swapped data if platform = XBOX360 else data
    :rtype: bytes
    """
    if (
        build_target == BuildTarget.XBOX360
    ):  # swap bytes for Xbox confirmed,PS3 not encountered
        for i in range(0, len(image_data), 2):
            image_data[i : i + 2] = image_data[i : i + 2][::-1]
    return image_data


def pillow(
    image_data: bytes, width: int, height: int, mode: str, codec: str, args, swap=False
) -> Image:
    img = (
        Image.frombytes(mode, (width, height), image_data, codec, args)
        if width
        else Image.new(mode, (width, height))
    )
    if swap:
        channels = img.split()
        img = Image.merge(mode, [channels[x] for x in swap])
    return img


def atc(image_data: bytes, width: int, height: int, alpha: bool) -> Image:
    if alpha:
        image_data = texture2ddecoder.decode_atc_rgba8(image_data, width, height, alpha)
    else:
        image_data = texture2ddecoder.decode_atc_rgb4(image_data, width, height)

    return Image.frombytes("RGBA", (width, height), image_data, "raw", "BGRA")


def astc(image_data: bytes, width: int, height: int, block_size: tuple) -> Image:
    image_data = texture2ddecoder.decode_astc(image_data, width, height, *block_size)
    return Image.frombytes("RGBA", (width, height), image_data, "raw", "BGRA")


def pvrtc(image_data: bytes, width: int, height: int, fmt: bool):
    image_data = texture2ddecoder.decode_pvrtc(image_data, width, height, fmt)
    return Image.frombytes("RGBA", (width, height), image_data, "raw", "BGRA")


def etc(image_data: bytes, width: int, height: int, fmt: list):
    if fmt[0] == 1:
        image_data = texture2ddecoder.decode_etc1(image_data, width, height)
    elif fmt[0] == 2:
        if fmt[1] == "RGB":
            image_data = texture2ddecoder.decode_etc2(image_data, width, height)
        elif fmt[1] == "A1":
            image_data = texture2ddecoder.decode_etc2a1(image_data, width, height)
        elif fmt[1] == "A8":
            image_data = texture2ddecoder.decode_etc2a8(image_data, width, height)
    else:
        raise NotImplementedError("unknown etc mode")
    return Image.frombytes("RGBA", (width, height), image_data, "raw", "BGRA")

def eac(image_data: bytes, width: int, height: int, fmt: list):
    if fmt == "EAC_R":
        image_data = texture2ddecoder.decode_eacr(image_data, width, height)
    elif fmt == "EAC_R_SIGNED":
        image_data = texture2ddecoder.decode_eacr_signed(image_data, width, height)
    elif fmt == "EAC_RG":
        image_data = texture2ddecoder.decode_eacrg(image_data, width, height)
    elif fmt == "EAC_RG_SIGNED":
        image_data = texture2ddecoder.decode_eacrg_signed(image_data, width, height)
    return Image.frombytes("RGBA", (width, height), image_data, "raw", "BGRA")



CONV_TABLE = {
#  FORMAT                  FUNC     #ARGS.....
#----------------------- -------- -------- ------------ ----------------- ------------ ----------
(  TF.Alpha8,              pillow,  "RGBA",  "raw",       "A"                                   ),
(  TF.ARGB4444,            pillow,  "RGBA",  "raw",       "RGBA;4B",          (2,1,0,3)         ),
(  TF.RGB24,               pillow,  "RGB",   "raw",       "RGB"                                 ),
(  TF.RGBA32,              pillow,  "RGBA",  "raw",       "RGBA"                                ),
(  TF.ARGB32,              pillow,  "RGBA",  "raw",       "ARGB"                                ),
(  TF.RGB565,              pillow,  "RGB",   "raw",       "BGR;16"                              ),
(  TF.R16,                 pillow,  "RGB",   "raw",       "R;16"                                ),
(  TF.DXT1,                pillow,  "RGBA",  "bcn",       1                                     ),
(  TF.DXT5,                pillow,  "RGBA",  "bcn",       3                                     ),
(  TF.RGBA4444,            pillow,  "RGBA",  "raw",       'RGBA;4B',          (3,2,1,0)         ),
(  TF.BGRA32,              pillow,  "RGBA",  "raw",       "BGRA"                                ),
(  TF.RHalf,                                                                                    ),
(  TF.RGHalf,                                                                                   ),
(  TF.RGBAHalf,                                                                                 ),
(  TF.RFloat,              pillow,  "RGB",   "raw",       "RF"                                  ),
(  TF.RGFloat,                                                                                  ),
(  TF.RGBAFloat,           pillow,  "RGBA",  "raw",       "RGBAF"                               ),
(  TF.YUY2,                                                                                     ),
(  TF.RGB9e5Float,                                                                              ),
(  TF.BC4,                 pillow,  "L",     "bcn",       4                                     ),
(  TF.BC5,                 pillow,  "RGBA",  "bcn",       5                                     ),
(  TF.BC6H,                pillow,  "RGBA",  "bcn",       6                                     ),
(  TF.BC7,                 pillow,  "RGBA",  "bcn",       7                                     ),
(  TF.DXT1Crunched,        pillow,  "RGBA",  "bcn",       1                                     ),
(  TF.DXT5Crunched,        pillow,  "RGBA",  "bcn",       3                                     ),
(  TF.PVRTC_RGB2,          pvrtc,   False                                                       ),
(  TF.PVRTC_RGBA2,         pvrtc,   True                                                        ),
(  TF.PVRTC_RGB4,          pvrtc,   False                                                       ),
(  TF.PVRTC_RGBA4,         pvrtc,   True                                                        ),
(  TF.ETC_RGB4,            etc,     (1,)                                                        ),
(  TF.ATC_RGB4,            atc,     False                                                       ),
(  TF.ATC_RGBA8,           atc,     True                                                        ),
(  TF.EAC_R,               eac,     "EAC_R"                                                     ),
(  TF.EAC_R_SIGNED,        eac,     "EAC_R:SIGNED"                                              ),
(  TF.EAC_RG,              eac,     "EAC_RG"                                                    ),
(  TF.EAC_RG_SIGNED,       eac,     "EAC_RG_SIGNED"                                             ),
(  TF.ETC2_RGB,            etc,     (2,"RGB")                                                   ),
(  TF.ETC2_RGBA1,          etc,     (2, "A1")                                                   ),
(  TF.ETC2_RGBA8,          etc,     (2, "A8")                                                   ),
(  TF.ASTC_RGB_4x4,        astc,    (4,4)                                                       ),
(  TF.ASTC_RGB_5x5,        astc,    (5,5)                                                       ),
(  TF.ASTC_RGB_6x6,        astc,    (6,6)                                                       ),
(  TF.ASTC_RGB_8x8,        astc,    (8,8)                                                       ),
(  TF.ASTC_RGB_10x10,      astc,    (10,10)                                                     ),
(  TF.ASTC_RGB_12x12,      astc,    (12,12)                                                     ),
(  TF.ASTC_RGBA_4x4,       astc,    (4,4)                                                       ),
(  TF.ASTC_RGBA_5x5,       astc,    (5,5)                                                       ),
(  TF.ASTC_RGBA_6x6,       astc,    (6,6)                                                       ),
(  TF.ASTC_RGBA_8x8,       astc,    (8,8)                                                       ),
(  TF.ASTC_RGBA_10x10,     astc,    (10,10)                                                     ),
(  TF.ASTC_RGBA_12x12,     astc,    (12,12)                                                     ),
(  TF.ETC_RGB4_3DS,        etc,     (1,)                                                        ),
(  TF.ETC_RGBA8_3DS,       etc,     (1,)                                                        ),
(  TF.RG16,                                                                                     ),
(  TF.R8,                  pillow,  "RGB",   "raw",       "R"                                   ),
(  TF.ETC_RGB4Crunched,    etc,     (1,)                                                        ),
(  TF.ETC2_RGBA8Crunched,  etc,     (2, "A8")                                                   ),
(  TF.ASTC_HDR_4x4,        astc,    (4,4)                                                       ),
(  TF.ASTC_HDR_5x5,        astc,    (5,5)                                                       ),
(  TF.ASTC_HDR_6x6,        astc,    (6,6)                                                       ),
(  TF.ASTC_HDR_8x8,        astc,    (8,8)                                                       ),
(  TF.ASTC_HDR_10x10,      astc,    (10,10)                                                     ),
(  TF.ASTC_HDR_12x12,      astc,    (12,12)                                                     ),
}

# format conv_table to a dict
CONV_TABLE = {line[0]: line[1:] for line in CONV_TABLE}

# XBOX Swap Formats
XBOX_SWAP_FORMATS = [TF.RGB565, TF.DXT1, TF.DXT1Crunched, TF.DXT5, TF.DXT5Crunched]



from .Audio import AudioType, AudioCompressionFormat, AUDIO_TYPE_EXTEMSION
from .BuildTarget import BuildTarget
from .ClassIDType import ClassIDType
from .FileType import FileType
from .TextureFormat import TextureFormat



class Vector2:
    X: float
    Y: float

    def __init__(self, x: float, y: float):
        self.X = x
        self.Y = y


"""
using System;
using System.Runtime.InteropServices;

namespace AssetStudio
{
    [StructLayout(LayoutKind.Sequential, Pack = 4)]
    public struct Vector2 : IEquatable<Vector2>
    {
        public float X;
        public float Y;

        public Vector2(float x, float y)
        {
            X = x;
            Y = y;
        }

        public float this[int index]
        {
            get
            {
                switch (index)
                {
                    case 0: return X;
                    case 1: return Y;
                    default: throw new ArgumentOutOfRangeException(nameof(index), "Invalid Vector2 index!");
                }
            }

            set
            {
                switch (index)
                {
                    case 0: X = value; break;
                    case 1: Y = value; break;
                    default: throw new ArgumentOutOfRangeException(nameof(index), "Invalid Vector2 index!");
                }
            }
        }

        public override int GetHashCode()
        {
            return X.GetHashCode() ^ (Y.GetHashCode() << 2);
        }

        public override bool Equals(object other)
        {
            if (!(other is Vector2))
                return false;
            return Equals((Vector2)other);
        }

        public bool Equals(Vector2 other)
        {
            return X.Equals(other.X) && Y.Equals(other.Y);
        }

        public void Normalize()
        {
            var length = Length();
            if (length > kEpsilon)
            {
                var invNorm = 1.0f / length;
                X *= invNorm;
                Y *= invNorm;
            }
            else
            {
                X = 0;
                Y = 0;
            }
        }

        public float Length()
        {
            return (float)Math.Sqrt(LengthSquared());
        }

        public float LengthSquared()
        {
            return X * X + Y * Y;
        }

        public static Vector2 Zero => new Vector2();

        public static Vector2 operator +(Vector2 a, Vector2 b)
        {
            return new Vector2(a.X + b.X, a.Y + b.Y);
        }

        public static Vector2 operator -(Vector2 a, Vector2 b)
        {
            return new Vector2(a.X - b.X, a.Y - b.Y);
        }

        public static Vector2 operator *(Vector2 a, Vector2 b)
        {
            return new Vector2(a.X * b.X, a.Y * b.Y);
        }

        public static Vector2 operator /(Vector2 a, Vector2 b)
        {
            return new Vector2(a.X / b.X, a.Y / b.Y);
        }

        public static Vector2 operator -(Vector2 a)
        {
            return new Vector2(-a.X, -a.Y);
        }

        public static Vector2 operator *(Vector2 a, float d)
        {
            return new Vector2(a.X * d, a.Y * d);
        }

        public static Vector2 operator *(float d, Vector2 a)
        {
            return new Vector2(a.X * d, a.Y * d);
        }

        public static Vector2 operator /(Vector2 a, float d)
        {
            return new Vector2(a.X / d, a.Y / d);
        }

        public static bool operator ==(Vector2 lhs, Vector2 rhs)
        {
            return (lhs - rhs).LengthSquared() < kEpsilon * kEpsilon;
        }

        public static bool operator !=(Vector2 lhs, Vector2 rhs)
        {
            return !(lhs == rhs);
        }

        public static implicit operator Vector3(Vector2 v)
        {
            return new Vector3(v.X, v.Y, 0);
        }

        public static implicit operator Vector4(Vector2 v)
        {
            return new Vector4(v.X, v.Y, 0.0F, 0.0F);
        }

        private const float kEpsilon = 0.00001F;
    }
}

"""



from .PPtr import PPtr
from .Texture import Texture


class MovieTexture(Texture):
    def __init__(self, reader):
        super().__init__(reader=reader)
        self.Loop = reader.read_boolean()
        reader.align_stream()
        self.AudioClip = PPtr(reader)  # AudioClip
        self.m_MovieData = reader.read_bytes(reader.read_int())



import os
import re

from .CompressionHelper import BROTLI_MAGIC, GZIP_MAGIC
from ..enums import FileType
from ..streams import EndianBinaryReader


def file_name_without_extension(file_name: str) -> str:
    return os.path.splitext(os.path.basename(file_name))[0]


def list_all_files(directory: str) -> list:
    return [
        val
        for sublist in [
            [os.path.join(dir_path, filename) for filename in filenames]
            for (dir_path, dirn_ames, filenames) in os.walk(directory)
            if ".git" not in dir_path
        ]
        for val in sublist
    ]


def find_all_files(directory: str, search_str: str) -> list:
    return [
        val
        for sublist in [
            [
                os.path.join(dir_path, filename)
                for filename in filenames
                if search_str in filename
            ]
            for (dir_path, dirn_ames, filenames) in os.walk(directory)
            if ".git" not in dir_path
        ]
        for val in sublist
    ]


def merge_split_assets(path: str, all_directories=False):
    if all_directories:
        split_files = [fp for fp in list_all_files(path) if fp[-7:] == ".split0"]
    else:
        split_files = [
            os.path.join(path, fp) for fp in os.listdir(path) if fp[-7:] == ".split0"
        ]

    for split_file in split_files:
        dest_file = file_name_without_extension(split_file)
        dest_path = os.path.dirname(split_file)
        dest_full = os.path.join(dest_file, dest_path)

        if not os.path.exists(dest_full):
            with open(dest_full, "wb") as f:
                i = 0
                while True:
                    split_part = "".join([dest_full, ".split", str(i)])
                    if not os.path.isfile(split_part):
                        break
                    f.write(open(split_part, "rb").read())


def processing_split_files(select_file: list) -> list:
    split_files = [fp for fp in select_file if ".split" in fp]
    select_file = [f for f in select_file if f not in split_files]

    split_files = set([file_name_without_extension(fp) for fp in split_files])
    for splitFile in split_files:
        if os.path.isfile:
            select_file.append(splitFile)
    return select_file


def check_file_type(input_) -> (FileType, EndianBinaryReader):
    if isinstance(input_, str) and os.path.isfile(input_):
        reader = EndianBinaryReader(open(input_, "rb"))
    elif isinstance(input_, EndianBinaryReader):
        reader = input_
    else:
        reader = EndianBinaryReader(input_)

    if reader.Length < 20:
        return FileType.ResourceFile, reader

    signature = reader.read_string_to_null(20)
    reader.Position = 0
    if signature in [
        "UnityWeb",
        "UnityRaw",
        "\xFA\xFA\xFA\xFA\xFA\xFA\xFA\xFA",
        "UnityFS",
    ]:
        return FileType.BundleFile, reader
    elif signature == "UnityWebData1.0":
        return FileType.WebFile, reader
    elif signature == "PK\x03\x04\x14":
        return FileType.ZIP, reader
    else:
        magic = reader.read_bytes(2)
        reader.Position = 0
        if GZIP_MAGIC == magic:
            return FileType.WebFile, reader
        reader.Position = 0x20
        magic = reader.read_bytes(6)
        reader.Position = 0
        if BROTLI_MAGIC == magic:
            return FileType.WebFile, reader

        # check if AssetsFile
        old_endian = reader.endian
        assets_file = True
        # read as if assetsfile and check version
        reader.Position = 0
        metadata_size = reader.read_u_int()
        file_size = reader.read_u_int()
        version = reader.read_u_int()
        data_offset = reader.read_u_int()

        if (
            version < 0
            or version > 100
            or any(
                [
                    x < 0 or x > reader.Length
                    for x in [metadata_size, version, data_offset]
                ]
            )
        ):
            return FileType.ResourceFile, reader

        if version >= 9:
            endian = ">" if reader.read_boolean() else "<"
            reserved = reader.read_bytes(3)
        else:
            reader.Position = file_size - metadata_size
            endian = ">" if reader.read_boolean() else "<"

        reader.endian = endian

        if version >= 7:
            unity_version = reader.read_string_to_null()
            if len([x for x in re.split(r"\D", unity_version) if x != ""]) != 4:
                assets_file = False

        # check end
        reader.endian = old_endian
        reader.Position = 0
        if assets_file:
            return FileType.AssetsFile, reader
        else:
            return FileType.ResourceFile, reader



from .NamedObject import NamedObject


class MonoScript(NamedObject):
    def __init__(self, reader):
        super().__init__(reader=reader)
        if self.version[0] > 3 or (
            self.version[0] == 3 and self.version[1] >= 4
        ):  # 3.4 and up
            self.execution_order = reader.read_int()
        if self.version[0] < 5:  # 5.0 down
            self.properties_hash = reader.read_u_int()
        else:
            self.properties_hash = reader.read_bytes(16)
        if self.version[0] < 3:  # 3.0 down
            self.path_name = reader.read_aligned_string()

        self.class_name = reader.read_aligned_string()
        if self.version[0] >= 3:  # 3.0 and up
            self.namespace = reader.read_aligned_string()
        self.assembly_name = reader.read_aligned_string()
        if self.version[0] < 2018 or (
            self.version[0] == 2018 and self.version[1] < 2
        ):  # 2018.2 down
            self.is_editor_script = reader.read_boolean()



import os
import re
import sys

from .File import File
from .. import classes
from ..CommonString import COMMON_STRING
from ..enums import BuildTarget, ClassIDType
from ..streams import EndianBinaryReader, EndianBinaryWriter

RECURSION_LIMIT = sys.getrecursionlimit()


class SerializedFileHeader:
    metadata_size: int
    file_size: int
    version: int
    data_offset: int
    endian: bytes
    reserved: bytes

    def __init__(self, reader: EndianBinaryReader):
        self.metadata_size = reader.read_u_int()
        self.file_size = reader.read_u_int()
        self.version = reader.read_u_int()
        self.data_offset = reader.read_u_int()


class LocalSerializedObjectIdentifier:  # script type
    local_serialized_file_index: int
    local_identifier_in_file: int

    def __init__(self, header: SerializedFileHeader, reader: EndianBinaryReader):
        self.local_serialized_file_index = reader.read_int()
        if header.version < 14:
            self.local_identifier_in_file = reader.read_int()
        else:
            reader.align_stream()
            self.local_identifier_in_file = reader.read_long()

    def write(self, header: SerializedFileHeader, writer: EndianBinaryWriter):
        writer.write_int(self.local_serialized_file_index)
        if header.version < 14:
            writer.write_int(self.local_identifier_in_file)
        else:
            writer.align_stream()
            writer.write_long(self.local_identifier_in_file)


class FileIdentifier:  # external
    guid: bytes
    type: int
    # enum { kNonAssetType = 0, kDeprecatedCachedAssetType = 1, kSerializedAssetType = 2, kMetaAssetType = 3 };
    path: str

    @property
    def name(self):
        return os.path.basename(self.path)

    def __init__(self, header: SerializedFileHeader, reader: EndianBinaryReader):
        if header.version >= 6:
            self.temp_empty = reader.read_string_to_null()
        if header.version >= 5:
            self.guid = reader.read_bytes(16)
            self.type = reader.read_int()
        self.path = reader.read_string_to_null()

    def write(self, header: SerializedFileHeader, writer: EndianBinaryWriter):
        if header.version >= 6:
            writer.write_string_to_null(self.temp_empty)
        if header.version >= 5:
            writer.write_bytes(self.guid)
            writer.write_int(self.type)
        writer.write_string_to_null(self.path)


class TypeTreeNode:
    type: str
    name: str
    byte_size: int
    index: int
    is_array: int
    version: int
    meta_flag: int
    level: int
    type_str_offset: int
    name_str_offset: int


class BuildType:
    build_type: str

    def __init__(self, build_type):
        self.build_type = build_type

    @property
    def is_alpha(self):
        return self.build_type == "a"

    @property
    def is_path(self):
        return self.build_type == "p"


class SerializedType:
    class_id: ClassIDType
    is_stripped_type: bool
    script_type_index = -1
    nodes: list = []  # TypeTreeNode
    script_id: bytes  # Hash128
    old_type_hash: bytes  # Hash128}


class SerializedFile(File):
    reader: EndianBinaryReader
    unity_version: str
    version: list
    build_type: BuildType
    target_platform: BuildTarget
    types: list
    script_types: list
    externals: list
    _container: dict
    objects: dict
    container: dict
    _cache: dict
    header: SerializedFileHeader

    @property
    def files(self):
        return self.objects

    def __init__(self, reader: EndianBinaryReader, environment=None, parent=None):
        self.environment = environment
        self.parent = parent
        self.reader = reader

        self.unity_version = "2.5.0f5"
        self.version = [0, 0, 0, 0]
        self.build_type = BuildType("")
        self.target_platform = BuildTarget.UnknownPlatform
        self._enable_type_tree = True
        self.types = []
        self.script_types = []
        self.externals = []
        self._container = {}

        self.objects = {}
        self.container = {}
        # used to speed up mass asset extraction
        # some assets refer to each other, so by keeping the result
        # of specific assets cached the extraction can be speed up by a lot.
        # used by: Sprite (Texture2D (with alpha) cached),
        self._cache = {}

        # ReadHeader
        header = SerializedFileHeader(reader)
        self.header = header

        if header.version >= 9:
            header.endian = ">" if reader.read_boolean() else "<"
            header.reserved = reader.read_bytes(3)
        else:
            reader.Position = header.file_size - header.metadata_size
            header.endian = ">" if reader.read_boolean() else "<"

        reader.endian = header.endian

        p1 = reader.Position
        if header.version >= 7:
            unity_version = reader.read_string_to_null()
            self.set_version(unity_version)

        if header.version >= 8:
            self._m_target_platform = reader.read_int()
            try:
                self.target_platform = BuildTarget(self._m_target_platform)
            except KeyError:
                self.target_platform = BuildTarget.UnknownPlatform

        if header.version >= 13:
            self._enable_type_tree = reader.read_boolean()

        # ReadTypes
        type_count = reader.read_int()
        self.types = [self.read_serialized_type() for _ in range(type_count)]

        if 7 <= header.version < 14:
            self.big_id_enabled = reader.read_int()

        # ReadObjects
        object_count = reader.read_int()
        self.objects = {}
        for _ in range(object_count):
            obj = ObjectReader(self, reader)
            self.objects[obj.path_id] = obj

        # Read Scripts
        if header.version >= 11:
            script_count = reader.read_int()
            self.script_types = [
                LocalSerializedObjectIdentifier(header, reader)
                for _ in range(script_count)
            ]

        # Read Externals
        externals_count = reader.read_int()
        self.externals = [
            FileIdentifier(header, reader) for _ in range(externals_count)
        ]

        self.userInformation = reader.read_string_to_null()
        if header.version >= 21:
            self.unknown = reader.ReadInt32()

        # read the asset_bundles to get the containers
        old_pos = reader.Position
        reader.Position = p1
        self.metadata = reader.read(old_pos - p1)
        reader.Position = old_pos
        for obj in self.objects.values():
            if obj.type == ClassIDType.AssetBundle:
                data = obj.read()
                for container, asset_info in data.container.items():
                    asset = asset_info.asset
                    self.container[container] = asset
                    if hasattr(asset, "path_id"):
                        self._container[asset.path_id] = container
        environment.container = {**environment.container, **self.container}

        reader.Position = old_pos

    def set_version(self, string_version):
        self.unity_version = string_version
        self.build_type = BuildType(re.findall(r"([^\d.])", string_version)[0])
        version_split = re.split(r"\D", string_version)
        self.version = [int(x) for x in version_split]

    def read_serialized_type(self):
        type_ = SerializedType()
        type_.class_id = self.reader.read_int()

        if self.header.version >= 16:
            type_.is_stripped_type = self.reader.read_boolean()

        if self.header.version >= 17:
            type_.script_type_index = self.reader.read_short()

        if self.header.version >= 13:
            if (self.header.version < 16 and type_.class_id < 0) or (
                self.header.version >= 16 and type_.class_id == 114
            ):
                type_.script_id = self.reader.read_bytes(16)  # Hash128
            type_.old_type_hash = self.reader.read_bytes(16)  # Hash128

        if self._enable_type_tree:
            type_tree = []
            if self.header.version >= 12 or self.header.version == 10:
                type_.string_data = self.read_type_tree5(type_tree)
            else:
                self.read_type_tree(type_tree)

            type_.nodes = type_tree

        return type_

    def read_type_tree(self, type_tree, level=0):
        if level == RECURSION_LIMIT - 1:
            raise RecursionError

        type_tree_node = TypeTreeNode()
        type_tree.append(type_tree_node)
        type_tree_node.level = level
        type_tree_node.type = self.reader.read_string_to_null()
        type_tree_node.name = self.reader.read_string_to_null()
        type_tree_node.byte_size = self.reader.read_int()
        if self.header.version == 2:
            type_tree_node.variable_count = self.reader.read_int()

        if self.header.version != 3:
            type_tree_node.index = self.reader.read_int()

        type_tree_node.is_array = self.reader.read_int()
        type_tree_node.version = self.reader.read_int()
        if self.header.version != 3:
            type_tree_node.meta_flag = self.reader.read_int()

        children_count = self.reader.read_int()
        for i in range(children_count):
            self.read_type_tree(type_tree, level + 1)

    def read_type_tree5(self, type_tree):
        number_of_nodes = self.reader.read_int()
        string_buffer_size = self.reader.read_int()

        node_size = 24
        if self.header.version > 17:
            node_size = 32

        self.reader.Position += number_of_nodes * node_size
        string_buffer_reader = EndianBinaryReader(self.reader.read(string_buffer_size))
        self.reader.Position -= number_of_nodes * node_size + string_buffer_size

        for i in range(number_of_nodes):
            type_tree_node = TypeTreeNode()
            type_tree.append(type_tree_node)
            type_tree_node.version = self.reader.read_u_short()
            type_tree_node.level = self.reader.read_byte()
            type_tree_node.is_array = self.reader.read_boolean()
            type_tree_node.type_str_offset = self.reader.read_u_int()
            type_tree_node.name_str_offset = self.reader.read_u_int()
            type_tree_node.byte_size = self.reader.read_int()
            type_tree_node.index = self.reader.read_int()
            type_tree_node.meta_flag = self.reader.read_int()

            if self.header.version > 17:
                type_tree_node.extra = self.reader.read(8)

            type_tree_node.type = read_string(
                string_buffer_reader, type_tree_node.type_str_offset
            )
            type_tree_node.name = read_string(
                string_buffer_reader, type_tree_node.name_str_offset
            )

        self.reader.Position += string_buffer_size
        if self.header.version >= 21:
            self.reader.Position += 4
        return string_buffer_reader.bytes

    def save(self) -> bytes:
        # Structure:
        #   1. header
        #       file header
        #       types
        #       objects
        #       scripts
        #       externals
        #   2. small 0es offset - align stream
        #   3. objects data

        # adjust metadata_size, data_offst and file_size
        header = self.header
        types = self.types
        objects = self.objects
        script_types = self.script_types
        externals = self.externals

        # the data-offset is required for the header and is kinda hard to calculate,
        # so we split the asset building up into multiple parts
        fileheader_writer = EndianBinaryWriter()
        header_writer = EndianBinaryWriter(endian=header.endian)
        data_writer = EndianBinaryWriter(endian=header.endian)

        # 1. building the header without file header

        # reader.Position = header.file_size - header.metadata_size
        # header.endian = '>' if reader.read_boolean() else '<'
        if header.version < 9:
            header_writer.write_boolean(header.endian == ">")

        if header.version >= 7:
            header_writer.write_string_to_null(self.unity_version)

        if header.version >= 8:
            header_writer.write_int(self._m_target_platform)

        if header.version >= 13:
            header_writer.write_boolean(self._enable_type_tree)

        # types
        header_writer.write_int(len(types))
        for typ in types:
            self.save_serialized_type(typ, header, header_writer)

        if 7 <= header.version < 14:
            header_writer.write_int(self.big_id_enabled)

        # objects
        header_writer.write_int(len(objects))
        for i, obj in enumerate(objects.values()):
            obj.write(header, header_writer, data_writer)
            if i < len(objects) - 1:
                data_writer.align_stream(8)

        # scripts
        if header.version >= 11:
            header_writer.write_int(len(script_types))
            for script_type in script_types:
                script_type.write(header, header_writer)

        # externals
        header_writer.write_int(len(externals))
        for external in externals:
            external.write(header, header_writer)

        header_writer.write_string_to_null(self.userInformation)
        if header.version >= 21:
            header_writer.write_int(self.unknown)

        if header.version >= 9:
            # file header
            fileheader_size = 4 * 4 + 1 + 3  # following + endian + reserved
            metadata_size = header_writer.Length
            # metadata size
            fileheader_writer.write_u_int(header_writer.Length)
            # align between metadata and data
            mod = (fileheader_size + metadata_size) % 16
            align_length = 16 - mod if mod else 0

            # file size
            fileheader_writer.write_u_int(
                header_writer.Length
                + data_writer.Length
                + fileheader_size
                + align_length
            )
            # version
            fileheader_writer.write_u_int(header.version)
            # data offset
            fileheader_writer.write_u_int(
                header_writer.Length + fileheader_size + align_length
            )
            # endian
            fileheader_writer.write_boolean(header.endian == ">")
            fileheader_writer.write_bytes(header.reserved)

            # self.reader.bytes[fileheader_size:fileheader_size+header_writer.Length] == header_writer.bytes

            return (
                fileheader_writer.bytes
                + header_writer.bytes
                + b"\x00" * align_length
                + data_writer.bytes
            )

        else:
            # file header
            fileheader_size = 4 * 4
            metadata_size = header_writer.Length
            # metadata size
            fileheader_writer.write_u_int(metadata_size)
            # file size
            fileheader_writer.write_u_int(
                metadata_size + data_writer.Length + fileheader_size
            )
            # version
            fileheader_writer.write_u_int(header.version)
            # data offset - unity seems to align the stream .... but it's apparently not necessary
            fileheader_writer.write_u_int(fileheader_size)

            return fileheader_writer.bytes + data_writer.bytes + header_writer.bytes

    def save_serialized_type(
        self,
        typ: SerializedType,
        header: SerializedFileHeader,
        writer: EndianBinaryWriter,
    ):
        writer.write_int(typ.class_id)

        if header.version >= 16:
            writer.write_boolean(typ.is_stripped_type)

        if header.version >= 17:
            writer.write_short(typ.script_type_index)

        if header.version >= 13:
            if (header.version < 16 and typ.class_id < 0) or (
                header.version >= 16 and typ.class_id == 114
            ):
                writer.write_bytes(typ.script_id)  # Hash128
            writer.write_bytes(typ.old_type_hash)  # Hash128

        if self._enable_type_tree:
            if header.version >= 12 or header.version == 10:
                self.save_type_tree5(typ.nodes, writer, typ.string_data)
            else:
                self.save_type_tree(typ.nodes, writer)

    def save_type_tree(self, nodes: list, writer: EndianBinaryWriter):
        for i, node in nodes:
            writer.write_string_to_null(node.type)
            writer.write_string_to_null(node.name)
            writer.write_int(node.byte_size)
            if self.header.version == 2:
                writer.write_int(node.variable_count)

            if self.header.version != 3:
                writer.write_int(node.index)

            writer.write_int(node.is_array)
            writer.write_int(node.version)
            if self.header.version != 3:
                writer.write_int(node.meta_flag)

            # calc children count
            children_count = 0
            for node2 in nodes[i + 1 :]:
                if node2.level == node.level:
                    break
                if node2.level == node.level - 1:
                    children_count += 1
            writer.write_int(children_count)

    def save_type_tree5(self, nodes: list, writer: EndianBinaryWriter, str_data=b""):
        # node count
        # stream buffer size
        # node data
        # string buffer

        string_buffer = EndianBinaryWriter()
        """
        saved_strings = {}
        def save_string(value: str) -> int:
            import re
            if value in saved_strings:
                return saved_strings[value]

            for key, val in COMMON_STRING.items():
                if val == value:
                    return key | 0x80000000
            if re.match(r"\d+", value):
                return int(value)
            else:
                # normal string
                offset = string_buffer.Position
                saved_strings[value] = offset
                string_buffer.write_string_to_null(value)
                return offset
        """
        string_buffer.write(str_data)
        strings_values = [
            (node.type_str_offset, node.name_str_offset) for node in nodes
        ]

        # number of nodes
        writer.write_int(len(nodes))
        # string buffer size
        writer.write_int(string_buffer.Length)

        # nodes
        for i, node in enumerate(nodes):
            # version
            writer.write_u_short(node.version)
            # level
            writer.write_byte(node.level)
            # is array
            writer.write_boolean(node.is_array)

            writer.write_u_int(strings_values[i][0])
            writer.write_u_int(strings_values[i][1])

            writer.write_int(node.byte_size)
            writer.write_int(node.index)
            writer.write_int(node.meta_flag)

            if self.header.version > 17:
                writer.write(node.extra)

        # string buffer
        writer.write(string_buffer.bytes)

        if self.header.version >= 21:
            self.reader.write_bytes(b"\x00" * 4)


def read_string(string_buffer_reader: EndianBinaryReader, value: int) -> str:
    is_offset = (value & 0x80000000) == 0
    if is_offset:
        string_buffer_reader.Position = value
        return string_buffer_reader.read_string_to_null()

    offset = value & 0x7FFFFFFF
    if offset in COMMON_STRING:
        return COMMON_STRING[offset]

    return str(offset)


class ObjectReader:
    byte_start: int
    byte_size: int
    type_id: int
    class_id: ClassIDType
    path_id: int
    serialized_type: SerializedType

    def __init__(self, assets_file, reader):
        self.assets_file = assets_file
        self.reader = reader
        self.data = b""
        self.version = assets_file.version
        self.version2 = assets_file.header.version
        self.platform = assets_file.target_platform
        self.build_type = assets_file.build_type

        header = assets_file.header
        types = assets_file.types

        if header.version < 14:
            self.path_id = reader.read_int()
        else:
            reader.align_stream()
            self.path_id = reader.read_long()

        self.byte_start = reader.read_u_int()
        self.byte_start += header.data_offset
        self.byte_size = reader.read_u_int()
        self.type_id = reader.read_int()
        if header.version < 16:
            self.class_id = reader.read_u_short()
            if types:
                self.serialized_type = [x for x in types if x.class_id == self.type_id][
                    0
                ]
            else:
                self.serialized_type = SerializedType()
            self.is_destroyed = reader.read_u_short()
        else:
            typ = types[self.type_id]
            self.serialized_type = typ
            self.class_id = typ.class_id

        self.type = ClassIDType(self.class_id)

        if header.version == 15 or header.version == 16:
            self.stripped = reader.read_byte()

    def write(self, header, writer, data_writer):
        # validated
        if header.version < 14:
            writer.write_int(self.path_id)
        else:
            writer.align_stream()
            writer.write_long(self.path_id)

        if self.data:
            data = self.data
        else:
            self.reset()
            data = self.reader.read(self.byte_size)
        writer.write_u_int(data_writer.Position)
        writer.write_u_int(len(data))

        data_writer.write(data)

        writer.write_int(self.type_id)

        if header.version < 16:
            # WARNING - CLASSIDTYPE MIGHT CHANGE THE NUMBER IF IT'S UNNOWN
            writer.write_u_short(self.class_id)
            writer.write_u_short(self.is_destroyed)

        if header.version == 15 or header.version == 16:
            writer.write_byte(self.stripped)

    @property
    def container(self):
        return (
            self.assets_file._container[self.path_id]
            if self.path_id in self.assets_file._container
            else None
        )

    def reset(self):
        self.reader.Position = self.byte_start

    def read(self):
        return getattr(classes, self.type.name, classes.Object)(self)

    def __getattr__(self, item: str):
        if hasattr(self.reader, item):
            return getattr(self.reader, item)

    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, self.type.name)



import gzip
import lzma
import struct

import brotli
import lz4.block

GZIP_MAGIC: bytes = b"\x1f\x8b"
BROTLI_MAGIC: bytes = b"brotli"


# LZMA
def decompress_lzma(data: bytes) -> bytes:
    """decompresses lzma-compressed data

	:param data: compressed data
	:type data: bytes
	:raises _lzma.LZMAError: Compressed data ended before the end-of-stream marker was reached
	:return: uncompressed data
	:rtype: bytes
	"""
    props, dict_size = struct.unpack("<BI", data[:5])
    lc = props % 9
    props = props // 9
    pb = props // 5
    lp = props % 5
    dec = lzma.LZMADecompressor(
        format=lzma.FORMAT_RAW,
        filters=[
            {
                "id": lzma.FILTER_LZMA1,
                "dict_size": dict_size,
                "lc": lc,
                "lp": lp,
                "pb": pb,
            }
        ],
    )
    return dec.decompress(data[5:])


def compress_lzma(data: bytes) -> bytes:
    """compresses data via lzma (unity specific)
	The current static settings may not be the best solution,
	but they are the most commonly used values and should therefore be enough for the time being.

	:param data: uncompressed data
	:type data: bytes
	:return: compressed data
	:rtype: bytes
	"""
    ec = lzma.LZMACompressor(
        format=lzma.FORMAT_RAW,
        filters=[
            {"id": lzma.FILTER_LZMA1, "dict_size": 524288, "lc": 3, "lp": 0, "pb": 2,}
        ],
    )
    ec.compress(data)
    return b"]\x00\x00\x08\x00" + ec.flush()


# LZ4
def decompress_lz4(data: bytes, uncompressed_size: int) -> bytes:  # LZ4M/LZ4HC
    """decompresses lz4-compressed data

	:param data: compressed data
	:type data: bytes
	:param uncompressed_size: size of the uncompressed data
	:type uncompressed_size: int
	:raises _block.LZ4BlockError: Decompression failed: corrupt input or insufficient space in destination buffer.
	:return: uncompressed data
	:rtype: bytes
	"""
    return lz4.block.decompress(data, uncompressed_size)


def compress_lz4(data: bytes) -> bytes:  # LZ4M/LZ4HC
    """compresses data via lz4.block

	:param data: uncompressed data
	:type data: bytes
	:return: compressed data
	:rtype: bytes
	"""
    return lz4.block.compress(
        data, mode="high_compression", compression=9, store_size=False
    )


# Brotli
def decompress_brotli(data: bytes) -> bytes:
    """decompresses brotli-compressed data

	:param data: compressed data
	:type data: bytes
	:raises brotli.error: BrotliDecompress failed
	:return: uncompressed data
	:rtype: bytes
	"""
    return brotli.decompress(data)


def compress_brotli(data: bytes) -> bytes:
    """compresses data via brotli

	:param data: uncompressed data
	:type data: bytes
	:return: compressed data
	:rtype: bytes
	"""
    return brotli.compress(data)


# GZIP
def decompress_gzip(data: bytes) -> bytes:
    """decompresses gzip-compressed data

	:param data: compressed data
	:type data: bytes
	:raises OSError: Not a gzipped file
	:return: uncompressed data
	:rtype: bytes
	"""
    return gzip.decompress(data)


def compress_gzip(data: bytes) -> bytes:
    """compresses data via gzip
	The current static settings may not be the best solution,
	but they are the most commonly used values and should therefore be enough for the time being.

	:param data: uncompressed data
	:type data: bytes
	:return: compressed data
	:rtype: bytes
	"""
    return gzip.compress(data)



class Rectangle:
    height: int
    width: int
    x: int
    y: int

    def __init__(self, *args, **kwargs):
        if args:
            # Rectangle(Point, Size)
            if len(args) == 4:
                self.x, self.y, self.width, self.height = args
        elif kwargs:
            self.__dict__.update(kwargs)

    def round(rect):
        return Rectangle(
            round(rect.x), round(rect.y), round(rect.width), round(rect.height)
        )

    @property
    def left(self):
        return self.x

    @property
    def top(self):
        return self.y

    @property
    def right(self):
        return self.x + self.width

    @property
    def bottom(self):
        return self.y + self.height

    @property
    def size(self):
        return (self.width, self.height)

    @property
    def location(self):
        return (self.x, self.y)



from .PPtr import PPtr
from .RuntimeAnimatorController import RuntimeAnimatorController


class AnimationClipOverride:
    def __init__(self, reader):
        self.original_clip = PPtr(reader)
        self.override_clip = PPtr(reader)


class AnimatorOverrideController(RuntimeAnimatorController):
    def __init__(self, reader):
        super().__init__(reader=reader)
        self.controller = PPtr(reader)
        num_overrides = reader.read_int()
        self.clips = [AnimationClipOverride(reader) for i in range(num_overrides)]



from .NamedObject import NamedObject
from .PPtr import PPtr


class AssetInfo:
    def __init__(self, reader):
        self.preload_index = reader.read_int()
        self.preload_size = reader.read_int()
        self.asset = PPtr(reader)


class AssetBundle(NamedObject):
    def __init__(self, reader):
        super().__init__(reader=reader)
        preload_table_size = reader.read_int()
        self.preload_table = [PPtr(reader) for _ in range(preload_table_size)]
        container_size = reader.read_int()
        self.container = {}
        for i in range(container_size):
            key = reader.read_aligned_string()
            self.container[key] = AssetInfo(reader)



from .PPtr import PPtr
from .Renderer import Renderer


class SkinnedMeshRenderer(Renderer):
    def __init__(self, reader):
        super().__init__(reader=reader)
        version = self.version
        self.m_Quality = reader.read_int()
        self.m_UpdateWhenOffscreen = reader.read_boolean()
        self.m_SkinNormals = reader.read_boolean()  # 3.1.0 and below
        reader.align_stream()

        if version[0] == 2 and version[1] < 6:  # 2.6 down
            self.m_DisableAnimationWhenOffscreen = PPtr(reader)  # Animation

        self.m_Mesh = PPtr(reader)  # Mesh

        m_BonesSize = reader.read_int()
        self.m_Bones = [PPtr(reader) for _ in range(m_BonesSize)]

        if version[0] > 4 or (version[0] == 4 and version[1] >= 3):  # 4.3 and up
            self.m_BlendShapeWeights = reader.read_float_array()



from .NamedObject import NamedObject
from .PPtr import PPtr


class Font(NamedObject):
    def __init__(self, reader):
        super().__init__(reader=reader)
        if (self.version[0] == 5 and self.version[1] >= 5) or self.version[
            0
        ] > 5:  # 5.5 and up:
            self.LineSpacing = reader.read_float()
            self.DefaultMaterial = PPtr(reader)
            self.FontSize = reader.read_float()
            self.Texture = PPtr(reader)
            self.AsciiStartOffset = reader.read_int()
            self.Tracking = reader.read_float()
            self.CharacterSpacing = reader.read_int()
            self.CharacterPadding = reader.read_int()
            self.ConvertCase = reader.read_int()
            CharacterRects_size = reader.read_int()
            for i in range(CharacterRects_size):
                reader.Position += 44  # CharacterInfo data 41
            KerningValues_size = reader.read_int()
            for i in range(KerningValues_size):
                reader.Position += 8
            self.PixelScale = reader.read_float()
            FontData_size = reader.read_int()
            if FontData_size > 0:
                self.FontData = reader.read_bytes(FontData_size)
        else:
            self.AsciiStartOffset = reader.read_int()

            if self.version[0] <= 3:
                self.FontCountX = reader.read_int()
                self.FontCountY = reader.read_int()

            self.Kerning = reader.read_float()
            self.LineSpacing = reader.read_float()

            if self.version[0] <= 3:
                PerCharacterKerning_size = reader.read_int()
                for i in range(PerCharacterKerning_size):
                    first = reader.read_int()
                    second = reader.read_float()
            else:
                self.CharacterSpacing = reader.read_int()
                self.CharacterPadding = reader.read_int()

            self.ConvertCase = reader.read_int()
            self.DefaultMaterial = PPtr(reader)

            CharacterRects_size = reader.read_int()
            for i in range(CharacterRects_size):
                index = reader.read_int()
                # Rectf uv
                uvx = reader.read_float()
                uvy = reader.read_float()
                uvwidth = reader.read_float()
                uvheight = reader.read_float()
                # Rectf vert
                vertx = reader.read_float()
                verty = reader.read_float()
                vertwidth = reader.read_float()
                vertheight = reader.read_float()
                width = reader.read_float()

                if self.version[0] >= 4:
                    flipped = reader.read_boolean()
                    reader.align_stream()

            self.Texture = PPtr(reader)

            KerningValues_size = reader.read_int()
            for i in range(KerningValues_size):
                pairfirst = reader.read_short()
                pairsecond = reader.read_short()
                second = reader.read_float()

            if self.version[0] <= 3:
                self.GridFont = reader.read_boolean()
                reader.align_stream()
            else:
                self.PixelScale = reader.read_float()
            FontData_size = reader.read_int()
            if FontData_size > 0:
                self.FontData = reader.read_bytes(FontData_size)



class Matrix4x4:
    M: list

    def __init__(self, values):
        if len(values) != 16:
            raise ValueError(
                "There must be sixteen and only sixteen input values for Matrix."
            )
        self.M = [values[i * 4 : i * 4 + 4] for i in range(0, 4)]


"""
using System;
using System.Runtime.InteropServices;

namespace AssetStudio
{
    [StructLayout(LayoutKind.Sequential, Pack = 4)]
    public struct Matrix4x4 : IEquatable<Matrix4x4>
    {
        public float M00;
        public float M10;
        public float M20;
        public float M30;

        public float M01;
        public float M11;
        public float M21;
        public float M31;

        public float M02;
        public float M12;
        public float M22;
        public float M32;

        public float M03;
        public float M13;
        public float M23;
        public float M33;

        public Matrix4x4(float[] values)
        {
            if (values == null)
                throw new ArgumentNullException(nameof(values));
            if (values.Length != 16)
                throw new ArgumentOutOfRangeException(nameof(values), "There must be sixteen and only sixteen input values for Matrix.");

            M00 = values[0];
            M10 = values[1];
            M20 = values[2];
            M30 = values[3];

            M01 = values[4];
            M11 = values[5];
            M21 = values[6];
            M31 = values[7];

            M02 = values[8];
            M12 = values[9];
            M22 = values[10];
            M32 = values[11];

            M03 = values[12];
            M13 = values[13];
            M23 = values[14];
            M33 = values[15];
        }

        public float this[int row, int column]
        {
            get => this[row + column * 4];

            set => this[row + column * 4] = value;
        }

        public float this[int index]
        {
            get
            {
                switch (index)
                {
                    case 0: return M00;
                    case 1: return M10;
                    case 2: return M20;
                    case 3: return M30;
                    case 4: return M01;
                    case 5: return M11;
                    case 6: return M21;
                    case 7: return M31;
                    case 8: return M02;
                    case 9: return M12;
                    case 10: return M22;
                    case 11: return M32;
                    case 12: return M03;
                    case 13: return M13;
                    case 14: return M23;
                    case 15: return M33;
                    default: throw new ArgumentOutOfRangeException(nameof(index), "Invalid Matrix4x4 index!");
                }
            }

            set
            {
                switch (index)
                {
                    case 0: M00 = value; break;
                    case 1: M10 = value; break;
                    case 2: M20 = value; break;
                    case 3: M30 = value; break;
                    case 4: M01 = value; break;
                    case 5: M11 = value; break;
                    case 6: M21 = value; break;
                    case 7: M31 = value; break;
                    case 8: M02 = value; break;
                    case 9: M12 = value; break;
                    case 10: M22 = value; break;
                    case 11: M32 = value; break;
                    case 12: M03 = value; break;
                    case 13: M13 = value; break;
                    case 14: M23 = value; break;
                    case 15: M33 = value; break;
                    default: throw new ArgumentOutOfRangeException(nameof(index), "Invalid Matrix4x4 index!");
                }
            }
        }

        public override int GetHashCode()
        {
            return GetColumn(0).GetHashCode() ^ (GetColumn(1).GetHashCode() << 2) ^ (GetColumn(2).GetHashCode() >> 2) ^ (GetColumn(3).GetHashCode() >> 1);
        }

        public override bool Equals(object other)
        {
            if (!(other is Matrix4x4))
                return false;
            return Equals((Matrix4x4)other);
        }

        public bool Equals(Matrix4x4 other)
        {
            return GetColumn(0).Equals(other.GetColumn(0))
                   && GetColumn(1).Equals(other.GetColumn(1))
                   && GetColumn(2).Equals(other.GetColumn(2))
                   && GetColumn(3).Equals(other.GetColumn(3));
        }

        public Vector4 GetColumn(int index)
        {
            switch (index)
            {
                case 0: return new Vector4(M00, M10, M20, M30);
                case 1: return new Vector4(M01, M11, M21, M31);
                case 2: return new Vector4(M02, M12, M22, M32);
                case 3: return new Vector4(M03, M13, M23, M33);
                default: throw new IndexOutOfRangeException("Invalid column index!");
            }
        }

        public Vector4 GetRow(int index)
        {
            switch (index)
            {
                case 0: return new Vector4(M00, M01, M02, M03);
                case 1: return new Vector4(M10, M11, M12, M13);
                case 2: return new Vector4(M20, M21, M22, M23);
                case 3: return new Vector4(M30, M31, M32, M33);
                default: throw new IndexOutOfRangeException("Invalid row index!");
            }
        }

        public static Matrix4x4 operator *(Matrix4x4 lhs, Matrix4x4 rhs)
        {
            Matrix4x4 res;
            res.M00 = lhs.M00 * rhs.M00 + lhs.M01 * rhs.M10 + lhs.M02 * rhs.M20 + lhs.M03 * rhs.M30;
            res.M01 = lhs.M00 * rhs.M01 + lhs.M01 * rhs.M11 + lhs.M02 * rhs.M21 + lhs.M03 * rhs.M31;
            res.M02 = lhs.M00 * rhs.M02 + lhs.M01 * rhs.M12 + lhs.M02 * rhs.M22 + lhs.M03 * rhs.M32;
            res.M03 = lhs.M00 * rhs.M03 + lhs.M01 * rhs.M13 + lhs.M02 * rhs.M23 + lhs.M03 * rhs.M33;

            res.M10 = lhs.M10 * rhs.M00 + lhs.M11 * rhs.M10 + lhs.M12 * rhs.M20 + lhs.M13 * rhs.M30;
            res.M11 = lhs.M10 * rhs.M01 + lhs.M11 * rhs.M11 + lhs.M12 * rhs.M21 + lhs.M13 * rhs.M31;
            res.M12 = lhs.M10 * rhs.M02 + lhs.M11 * rhs.M12 + lhs.M12 * rhs.M22 + lhs.M13 * rhs.M32;
            res.M13 = lhs.M10 * rhs.M03 + lhs.M11 * rhs.M13 + lhs.M12 * rhs.M23 + lhs.M13 * rhs.M33;

            res.M20 = lhs.M20 * rhs.M00 + lhs.M21 * rhs.M10 + lhs.M22 * rhs.M20 + lhs.M23 * rhs.M30;
            res.M21 = lhs.M20 * rhs.M01 + lhs.M21 * rhs.M11 + lhs.M22 * rhs.M21 + lhs.M23 * rhs.M31;
            res.M22 = lhs.M20 * rhs.M02 + lhs.M21 * rhs.M12 + lhs.M22 * rhs.M22 + lhs.M23 * rhs.M32;
            res.M23 = lhs.M20 * rhs.M03 + lhs.M21 * rhs.M13 + lhs.M22 * rhs.M23 + lhs.M23 * rhs.M33;

            res.M30 = lhs.M30 * rhs.M00 + lhs.M31 * rhs.M10 + lhs.M32 * rhs.M20 + lhs.M33 * rhs.M30;
            res.M31 = lhs.M30 * rhs.M01 + lhs.M31 * rhs.M11 + lhs.M32 * rhs.M21 + lhs.M33 * rhs.M31;
            res.M32 = lhs.M30 * rhs.M02 + lhs.M31 * rhs.M12 + lhs.M32 * rhs.M22 + lhs.M33 * rhs.M32;
            res.M33 = lhs.M30 * rhs.M03 + lhs.M31 * rhs.M13 + lhs.M32 * rhs.M23 + lhs.M33 * rhs.M33;

            return res;
        }

        public static bool operator ==(Matrix4x4 lhs, Matrix4x4 rhs)
        {
            return lhs.GetColumn(0) == rhs.GetColumn(0)
                && lhs.GetColumn(1) == rhs.GetColumn(1)
                && lhs.GetColumn(2) == rhs.GetColumn(2)
                && lhs.GetColumn(3) == rhs.GetColumn(3);
        }

        public static bool operator !=(Matrix4x4 lhs, Matrix4x4 rhs)
        {
            return !(lhs == rhs);
        }

        public static Matrix4x4 Scale(Vector3 vector)
        {
            Matrix4x4 m;
            m.M00 = vector.X; m.M01 = 0F; m.M02 = 0F; m.M03 = 0F;
            m.M10 = 0F; m.M11 = vector.Y; m.M12 = 0F; m.M13 = 0F;
            m.M20 = 0F; m.M21 = 0F; m.M22 = vector.Z; m.M23 = 0F;
            m.M30 = 0F; m.M31 = 0F; m.M32 = 0F; m.M33 = 1F;
            return m;
        }

        public static Matrix4x4 Translate(Vector3 vector)
        {
            Matrix4x4 m;
            m.M00 = 1F; m.M01 = 0F; m.M02 = 0F; m.M03 = vector.X;
            m.M10 = 0F; m.M11 = 1F; m.M12 = 0F; m.M13 = vector.Y;
            m.M20 = 0F; m.M21 = 0F; m.M22 = 1F; m.M23 = vector.Z;
            m.M30 = 0F; m.M31 = 0F; m.M32 = 0F; m.M33 = 1F;
            return m;
        }

        public static Matrix4x4 Rotate(Quaternion q)
        {
            float x = q.X * 2.0F;
            float y = q.Y * 2.0F;
            float z = q.Z * 2.0F;
            float xx = q.X * x;
            float yy = q.Y * y;
            float zz = q.Z * z;
            float xy = q.X * y;
            float xz = q.X * z;
            float yz = q.Y * z;
            float wx = q.W * x;
            float wy = q.W * y;
            float wz = q.W * z;

            Matrix4x4 m;
            m.M00 = 1.0f - (yy + zz); m.M10 = xy + wz; m.M20 = xz - wy; m.M30 = 0.0F;
            m.M01 = xy - wz; m.M11 = 1.0f - (xx + zz); m.M21 = yz + wx; m.M31 = 0.0F;
            m.M02 = xz + wy; m.M12 = yz - wx; m.M22 = 1.0f - (xx + yy); m.M32 = 0.0F;
            m.M03 = 0.0F; m.M13 = 0.0F; m.M23 = 0.0F; m.M33 = 1.0F;
            return m;
        }
    }
}

"""



from .Component import Component
from .PPtr import PPtr


class StaticBatchInfo:
    def __init__(self, reader):
        self.firstSubMesh = reader.read_u_short()
        self.subMeshCount = reader.read_u_short()


class Renderer(Component):
    def __init__(self, reader):
        super().__init__(reader=reader)
        version = self.version
        if version[0] < 5:  # 5.0 down
            self.m_Enabled = reader.read_boolean()
            self.m_CastShadows = reader.read_boolean()
            self.m_ReceiveShadows = reader.read_boolean()
            self.m_LightmapIndex = reader.read_byte()
        else:  # 5.0 and up
            if version[0] > 5 or (version[0] == 5 and version[1] >= 4):  # 5.4 and up
                self.m_Enabled = reader.read_boolean()
                self.m_CastShadows = reader.read_byte()
                self.m_ReceiveShadows = reader.read_byte()
                # 2017.2 and up
                if version[0] > 2017 or (version[0] == 2017 and version[0] >= 2):
                    self.m_DynamicOccludee = reader.read_byte()
                self.m_MotionVectors = reader.read_byte()
                self.m_LightProbeUsage = reader.read_byte()
                self.m_ReflectionProbeUsage = reader.read_byte()
                # 2019.3 and up
                if version[0] > 2019 or (version[0] == 2019 and version[0] >= 3):
                    self.m_RayTracingMode = reader.ReadByte()
                reader.align_stream()
            else:
                self.m_Enabled = reader.read_boolean()
                reader.align_stream()
                self.m_CastShadows = reader.read_byte()
                self.m_ReceiveShadows = reader.read_boolean()
                reader.align_stream()

            if version[0] >= 2018:  # 2018 and up
                self.m_RenderingLayerMask = reader.read_u_int()

            # 2018.3 and up
            if version[0] > 2018 or (version[0] == 2018 and version[1] >= 3):
                self.m_RendererPriority = reader.read_int()

            self.m_LightmapIndex = reader.read_u_short()
            self.m_LightmapIndexDynamic = reader.read_u_short()

        if version[0] >= 3:  # 3.0 and up
            self.m_LightmapTilingOffset = reader.read_vector4()

        if version[0] >= 5:  # 5.0 and up
            self.m_LightmapTilingOffsetDynamic = reader.read_vector4()

        m_MaterialsSize = reader.read_int()
        self.m_Materials = [PPtr(reader) for _ in range(m_MaterialsSize)]  # Material

        if version[0] < 3:  # 3.0 down
            self.m_LightmapTilingOffset = reader.read_vector4()
        else:  # 3.0 and up
            if version[0] > 5 or (version[0] == 5 and version[1] >= 5):  # 5.5 and up
                self.m_StaticBatchInfo = StaticBatchInfo(reader)
            else:
                self.m_SubsetIndices = reader.read_u_int_array()

            self.m_StaticBatchRoot = PPtr(reader)  # Transform

        if version[0] > 5 or (version[0] == 5 and version[1] >= 4):  # 5.4 and up
            self.m_ProbeAnchor = PPtr(reader)  # Transform
            self.m_LightProbeVolumeOverride = PPtr(reader)  # GameObject
        elif version[0] > 3 or (version[0] == 3 and version[1] >= 5):  # 3.5 - 5.3
            self.m_UseLightProbes = reader.read_boolean()
            reader.align_stream()

            if version[0] >= 5:  # 5.0 and up
                self.m_ReflectionProbeUsage = reader.read_int()

            # Transform #5.0 and up m_ProbeAnchor
            self.m_LightProbeAnchor = PPtr(reader)

        if version[0] > 4 or (version[0] == 4 and version[1] >= 3):  # 4.3 and up
            if version[0] == 4 and version[1] == 3:  # 4.3
                self.m_SortingLayer = reader.read_short()
            else:
                self.m_SortingLayerID = reader.read_u_int()

            # SInt16 m_SortingLayer 5.6 and up
            self.m_SortingOrder = reader.read_short()
            reader.align_stream()



from .NamedObject import NamedObject
from ..streams import EndianBinaryWriter


class Texture(NamedObject):
    def __init__(self, reader):
        super().__init__(reader=reader)
        if self.version[0] > 2017 or (
            self.version[0] == 2017 and self.version[1] >= 3
        ):  # 2017.3 and up
            self.forced_fallback_format = reader.read_int()
            self.downscale_fallback = reader.read_boolean()
            reader.align_stream()

    def save(self, writer: EndianBinaryWriter):
        super().save(writer)
        if self.version[0] > 2017 or (
            self.version[0] == 2017 and self.version[1] >= 3
        ):  # 2017.3 and up
            writer.write_int(self.forced_fallback_format)
            writer.write_boolean(self.downscale_fallback)
            writer.align_stream()



import struct

from UnityPy import math

# import binascii

# import numpy

MaxValue = 65504.0
MinValue = -65504.0


def ToHalf(*args) -> float:
    """
	Converts the input into a half-float.
	Inputs:
		unsigned integer
		or
		buffer (bytes, buffer)
		offset
	"""
    # int input -> pack as UInt16
    if len(args) == 1:
        data = struct.pack("H", args[0])
        val = struct.unpack("e", data)[0]
    # buffer input
    elif len(args) == 2:
        val = struct.unpack_from("e", args[0], args[1])[0]

    if math.isnan(val):
        # print('Nan')
        return 0
    elif math.isinf(val):
        return MaxValue

    return val


# #CONSTANTS
# Epsilon = ToHalf(0x0001)
# MaxValue = ToHalf(0x7bff)
# MinValue = ToHalf(0xfbff)
# NaN = ToHalf(0xfe00)
# NegativeInfinity = ToHalf(0xfc00)
# PositiveInfinity = ToHalf(0x7c00)

# class Float16Compressor:
# 	def __init__(self):
# 		self.temp = 0

# 	def compress(self, float32):
# 		F16_EXPONENT_BITS = 0x1F
# 		F16_EXPONENT_SHIFT = 10
# 		F16_EXPONENT_BIAS = 15
# 		F16_MANTISSA_BITS = 0x3ff
# 		F16_MANTISSA_SHIFT = (23 - F16_EXPONENT_SHIFT)
# 		F16_MAX_EXPONENT = (F16_EXPONENT_BITS << F16_EXPONENT_SHIFT)

# 		a = struct.pack('>f', float32)
# 		b = binascii.hexlify(a)

# 		f32 = int(b, 16)
# 		f16 = 0
# 		sign = (f32 >> 16) & 0x8000
# 		exponent = ((f32 >> 23) & 0xff) - 127
# 		mantissa = f32 & 0x007fffff

# 		if exponent == 128:
# 			f16 = sign | F16_MAX_EXPONENT
# 			if mantissa:
# 				f16 |= (mantissa & F16_MANTISSA_BITS)
# 		elif exponent > 15:
# 			f16 = sign | F16_MAX_EXPONENT
# 		elif exponent > -15:
# 			exponent += F16_EXPONENT_BIAS
# 			mantissa >>= F16_MANTISSA_SHIFT
# 			f16 = sign | exponent << F16_EXPONENT_SHIFT | mantissa
# 		else:
# 			f16 = sign
# 		return f16

# 	def decompress(self, float16):
# 		s = int((float16 >> 15) & 0x00000001)  # sign
# 		e = int((float16 >> 10) & 0x0000001f)  # exponent
# 		f = int(float16 & 0x000003ff)  # fraction

# 		if e == 0:
# 			if f == 0:
# 				return int(s << 31)
# 			else:
# 				while not (f & 0x00000400):
# 					f = f << 1
# 					e -= 1
# 				e += 1
# 				f &= ~0x00000400
# 		# print(s,e,f)
# 		elif e == 31:
# 			if f == 0:
# 				return int((s << 31) | 0x7f800000)
# 			else:
# 				return int((s << 31) | 0x7f800000 | (f << 13))

# 		e = e + (127 - 15)
# 		f = f << 13
# 		return int((s << 31) | (e << 23) | f)



from .Behaviour import Behaviour
from .PPtr import PPtr


class Animation(Behaviour):
    def __init__(self, reader):
        super().__init__(reader=reader)
        self.animation = PPtr(reader)
        num_animations = reader.read_int()
        self.animations = [PPtr(reader) for i in range(num_animations)]



from .Texture import Texture
from ..enums import TextureFormat
from ..export import Texture2DConverter
from ..helpers.ResourceReader import get_resource_data
from ..streams import EndianBinaryWriter


class Texture2D(Texture):
    @property
    def image(self):
        return Texture2DConverter.get_image_from_texture2d(self)

    @image.setter
    def image(self, img):
        # img is PIL.Image
        # overwrite original image data with the RGB(A) image data and sets the correct new format
        writer = EndianBinaryWriter()
        for pix in img.getdata():
            for val in pix:
                writer.write_u_byte(val)
        self.image_data = writer.bytes
        self.m_TextureFormat = (
            TextureFormat.RGBA32 if len(pix) == 4 else TextureFormat.RGB24
        )

    def __init__(self, reader):
        super().__init__(reader=reader)
        version = self.version
        self.m_Width = reader.read_int()
        self.m_Height = reader.read_int()
        self.m_CompleteImageSize = reader.read_int()
        self.m_TextureFormat = TextureFormat(reader.read_int())
        if version[0] < 5 or (version[0] == 5 and version[1] < 2):  # 5.2 down
            self.m_MipMap = reader.read_boolean()
        else:
            self.m_MipCount = reader.read_int()
        self.m_IsReadable = reader.read_boolean()  # 2.6.0 and up
        self.m_ReadAllowed = reader.read_boolean()  # 3.0.0 - 5.4
        # bool m_StreamingMipmaps 2018.2 and up
        reader.align_stream()
        if version[0] > 2018 or (
            version[0] == 2018 and version[1] >= 2
        ):  # 2018.2 and up
            self.m_StreamingMipmapsPriority = reader.read_int()
        self.m_ImageCount = reader.read_int()
        self.m_TextureDimension = reader.read_int()
        self.m_TextureSettings = GLTextureSettings(reader)
        if version[0] >= 3:  # 3.0 and up
            self.m_LightmapFormat = reader.read_int()
        if version[0] > 3 or (version[0] == 3 and version[1] >= 5):  # 3.5.0 and up
            self.m_ColorSpace = reader.read_int()

        image_data_size = reader.read_int()
        self.image_data = b""
        if image_data_size == 0 and (
            (version[0] == 5 and version[1] >= 3) or version[0] > 5
        ):  # 5.3.0 and up
            m_StreamData = StreamingInfo(reader)

            self.image_data = get_resource_data(
                m_StreamData.path,
                self.assets_file,
                m_StreamData.offset,
                m_StreamData.size,
            )
        else:
            self.image_data = reader.read_bytes(image_data_size)

    def save(self, writer=None):
        if writer == None:
            writer = EndianBinaryWriter()
        version = self.version

        writer.write_int(self.m_Width)
        writer.write_int(self.m_Height)
        writer.write_int(self.m_CompleteImageSize)
        writer.write_int(int(self.m_TextureFormat))
        if version[0] < 5 or (version[0] == 5 and version[1] < 2):  # 5.2 down
            writer.write_boolean(self.m_MipMap)
        else:
            writer.write_int(self.m_MipCount)
        writer.write_boolean(self.m_IsReadable)  # 2.6.0 and up
        writer.write_boolean(self.m_ReadAllowed)  # 3.0.0 - 5.4
        # bool m_StreamingMipmaps 2018.2 and up
        writer.align_stream()
        if version[0] > 2018 or (
            version[0] == 2018 and version[1] >= 2
        ):  # 2018.2 and up
            writer.write_int(self.m_StreamingMipmapsPriority)
        writer.write_int(self.m_ImageCount)
        writer.write_int(self.m_TextureDimension)
        self.m_TextureSettings.save(writer, version)
        if version[0] >= 3:  # 3.0 and up
            writer.write_int(self.m_LightmapFormat)
        if version[0] > 3 or (version[0] == 3 and version[1] >= 5):  # 3.5.0 and up
            writer.write_int(self.m_ColorSpace)

        writer.write_int(len(self.image_data))
        writer.write_bytes(self.image_data)


class StreamingInfo:
    offset: int = 0
    size: int = 0
    path: str = ""

    def __init__(self, reader):
        self.offset = reader.read_u_int()
        self.size = reader.read_u_int()
        self.path = reader.read_aligned_string()


class GLTextureSettings:
    def __init__(self, reader):
        version = reader.version

        self.m_FilterMode = reader.read_int()
        self.m_Aniso = reader.read_int()
        self.m_MipBias = reader.read_float()
        if version[0] >= 2017:  # 2017.x and up
            self.m_WrapMode = reader.read_int()  # m_WrapU
            self.m_WrapV = reader.read_int()
            self.m_WrapW = reader.read_int()
        else:
            self.m_WrapMode = reader.read_int()

    def save(self, writer, version):
        writer.write_int(self.self.m_FilterMode)
        writer.write_int(self.self.m_Aniso)
        writer.write_float(self.self.m_MipBias)
        if version[0] >= 2017:  # 2017.x and up
            writer.write_int(self.self.m_WrapMode)  # m_WrapU
            writer.write_int(self.self.m_WrapV)
            writer.write_int(self.self.m_WrapW)
        else:
            writer.write_int(self.self.m_WrapMode)



import os

from .helpers import ImportHelper


class ResourceReader:
    need_search: bool
    path: str
    assets_file: None
    offset: int
    size: int

    def __init__(self, *args):
        """
		Input:
		Option 1:
			path - file path
			assets_file - SerializedFile
			offset -
			size -
		Option 2:
			reader - EndianBinaryReader
			offset -
			size -
		"""
        if len(args) == 4:
            self.need_search = True
            self.path = args[0]
            self.assets_file = args[1]
        elif len(args) == 3:
            self.need_search = False
            self.reader = args[0]
        self.offset = args[-2]
        self.size = args[-1]

    def get_data(self):
        if self.need_search:
            resource_file_name = os.path.basename(self.path)

            reader = self.assets_file.assets_manager.resource_file_readers.get(
                resource_file_name
            )
            if reader:
                reader.Position = self.offset
                return reader.read_bytes(self.size)

            current_directory = os.path.dirname(self.assets_file.full_name)
            resource_file_path = os.path.join(current_directory, resource_file_name)
            if not os.path.isfile(resource_file_path):
                find_files = ImportHelper.find_all_files(
                    current_directory, resource_file_name
                )
                if find_files:
                    resource_file_path = find_files[0]

            if os.path.isfile(resource_file_path):
                with open(resource_file_path, "rb") as f:
                    f.seek(self.offset)
                    return f.read(self.size)
            else:
                raise FileNotFoundError(
                    f"Can't find the resource file {resource_file_name}"
                )

        else:
            self.reader.Position = self.offset
            return self.reader.read_bytes(self.size)



from enum import IntEnum


class ClassIDType(IntEnum):
    UnknownType = (-1,)
    Object = (0,)
    GameObject = (1,)
    Component = (2,)
    LevelGameManager = (3,)
    Transform = (4,)
    TimeManager = (5,)
    GlobalGameManager = (6,)
    Behaviour = (8,)
    GameManager = (9,)
    AudioManager = (11,)
    ParticleAnimator = (12,)
    InputManager = (13,)
    EllipsoidParticleEmitter = (15,)
    Pipeline = (17,)
    EditorExtension = (18,)
    Physics2DSettings = (19,)
    Camera = (20,)
    Material = (21,)
    MeshRenderer = (23,)
    Renderer = (25,)
    ParticleRenderer = (26,)
    Texture = (27,)
    Texture2D = (28,)
    SceneSettings = (29,)
    GraphicsSettings = (30,)
    MeshFilter = (33,)
    OcclusionPortal = (41,)
    Mesh = (43,)
    Skybox = (45,)
    QualitySettings = (47,)
    Shader = (48,)
    TextAsset = (49,)
    Rigidbody2D = (50,)
    Physics2DManager = (51,)
    Collider2D = (53,)
    Rigidbody = (54,)
    PhysicsManager = (55,)
    Collider = (56,)
    Joint = (57,)
    CircleCollider2D = (58,)
    HingeJoint = (59,)
    PolygonCollider2D = (60,)
    BoxCollider2D = (61,)
    PhysicsMaterial2D = (62,)
    MeshCollider = (64,)
    BoxCollider = (65,)
    SpriteCollider2D = (66,)
    EdgeCollider2D = (68,)
    CapsuleCollider2D = (70,)
    ComputeShader = (72,)
    AnimationClip = (74,)
    ConstantForce = (75,)
    WorldParticleCollider = (76,)
    TagManager = (78,)
    AudioListener = (81,)
    AudioSource = (82,)
    AudioClip = (83,)
    RenderTexture = (84,)
    CustomRenderTexture = (86,)
    MeshParticleEmitter = (87,)
    ParticleEmitter = (88,)
    Cubemap = (89,)
    Avatar = (90,)
    AnimatorController = (91,)
    GUILayer = (92,)
    RuntimeAnimatorController = (93,)
    ScriptMapper = (94,)
    Animator = (95,)
    TrailRenderer = (96,)
    DelayedCallManager = (98,)
    TextMesh = (102,)
    RenderSettings = (104,)
    Light = (108,)
    CGProgram = (109,)
    BaseAnimationTrack = (110,)
    Animation = (111,)
    MonoBehaviour = (114,)
    MonoScript = (115,)
    MonoManager = (116,)
    Texture3D = (117,)
    NewAnimationTrack = (118,)
    Projector = (119,)
    LineRenderer = (120,)
    Flare = (121,)
    Halo = (122,)
    LensFlare = (123,)
    FlareLayer = (124,)
    HaloLayer = (125,)
    NavMeshAreas = (126,)
    HaloManager = (127,)
    Font = (128,)
    PlayerSettings = (129,)
    NamedObject = (130,)
    GUITexture = (131,)
    GUIText = (132,)
    GUIElement = (133,)
    PhysicMaterial = (134,)
    SphereCollider = (135,)
    CapsuleCollider = (136,)
    SkinnedMeshRenderer = (137,)
    FixedJoint = (138,)
    RaycastCollider = (140,)
    BuildSettings = (141,)
    AssetBundle = (142,)
    CharacterController = (143,)
    CharacterJoint = (144,)
    SpringJoint = (145,)
    WheelCollider = (146,)
    ResourceManager = (147,)
    NetworkView = (148,)
    NetworkManager = (149,)
    PreloadData = (150,)
    MovieTexture = (152,)
    ConfigurableJoint = (153,)
    TerrainCollider = (154,)
    MasterServerInterface = (155,)
    TerrainData = (156,)
    LightmapSettings = (157,)
    WebCamTexture = (158,)
    EditorSettings = (159,)
    InteractiveCloth = (160,)
    ClothRenderer = (161,)
    EditorUserSettings = (162,)
    SkinnedCloth = (163,)
    AudioReverbFilter = (164,)
    AudioHighPassFilter = (165,)
    AudioChorusFilter = (166,)
    AudioReverbZone = (167,)
    AudioEchoFilter = (168,)
    AudioLowPassFilter = (169,)
    AudioDistortionFilter = (170,)
    SparseTexture = (171,)
    AudioBehaviour = (180,)
    AudioFilter = (181,)
    WindZone = (182,)
    Cloth = (183,)
    SubstanceArchive = (184,)
    ProceduralMaterial = (185,)
    ProceduralTexture = (186,)
    Texture2DArray = (187,)
    CubemapArray = (188,)
    OffMeshLink = (191,)
    OcclusionArea = (192,)
    Tree = (193,)
    NavMeshObsolete = (194,)
    NavMeshAgent = (195,)
    NavMeshSettings = (196,)
    LightProbesLegacy = (197,)
    ParticleSystem = (198,)
    ParticleSystemRenderer = (199,)
    ShaderVariantCollection = (200,)
    LODGroup = (205,)
    BlendTree = (206,)
    Motion = (207,)
    NavMeshObstacle = (208,)
    TerrainInstance = (210,)
    SpriteRenderer = (212,)
    Sprite = (213,)
    CachedSpriteAtlas = (214,)
    ReflectionProbe = (215,)
    ReflectionProbes = (216,)
    Terrain = (218,)
    LightProbeGroup = (220,)
    AnimatorOverrideController = (221,)
    CanvasRenderer = (222,)
    Canvas = (223,)
    RectTransform = (224,)
    CanvasGroup = (225,)
    BillboardAsset = (226,)
    BillboardRenderer = (227,)
    SpeedTreeWindAsset = (228,)
    AnchoredJoint2D = (229,)
    Joint2D = (230,)
    SpringJoint2D = (231,)
    DistanceJoint2D = (232,)
    HingeJoint2D = (233,)
    SliderJoint2D = (234,)
    WheelJoint2D = (235,)
    ClusterInputManager = (236,)
    BaseVideoTexture = (237,)
    NavMeshData = (238,)
    AudioMixer = (240,)
    AudioMixerController = (241,)
    AudioMixerGroupController = (243,)
    AudioMixerEffectController = (244,)
    AudioMixerSnapshotController = (245,)
    PhysicsUpdateBehaviour2D = (246,)
    ConstantForce2D = (247,)
    Effector2D = (248,)
    AreaEffector2D = (249,)
    PointEffector2D = (250,)
    PlatformEffector2D = (251,)
    SurfaceEffector2D = (252,)
    BuoyancyEffector2D = (253,)
    RelativeJoint2D = (254,)
    FixedJoint2D = (255,)
    FrictionJoint2D = (256,)
    TargetJoint2D = (257,)
    LightProbes = (258,)
    LightProbeProxyVolume = (259,)
    SampleClip = (271,)
    AudioMixerSnapshot = (272,)
    AudioMixerGroup = (273,)
    NScreenBridge = (280,)
    AssetBundleManifest = (290,)
    UnityAdsManager = (292,)
    RuntimeInitializeOnLoadManager = (300,)
    CloudWebServicesManager = (301,)
    UnityAnalyticsManager = (303,)
    CrashReportManager = (304,)
    PerformanceReportingManager = (305,)
    UnityConnectSettings = (310,)
    AvatarMask = (319,)
    VideoPlayer = (328,)
    VideoClip = (329,)
    OcclusionCullingData = (363,)
    # kLargestRuntimeClassID = 364
    SmallestEditorClassID = (1000,)
    Prefab = (1001,)
    EditorExtensionImpl = (1002,)
    AssetImporter = (1003,)
    AssetDatabase = (1004,)
    Mesh3DSImporter = (1005,)
    TextureImporter = (1006,)
    ShaderImporter = (1007,)
    ComputeShaderImporter = (1008,)
    AudioImporter = (1020,)
    HierarchyState = (1026,)
    GUIDSerializer = (1027,)
    AssetMetaData = (1028,)
    DefaultAsset = (1029,)
    DefaultImporter = (1030,)
    TextScriptImporter = (1031,)
    SceneAsset = (1032,)
    NativeFormatImporter = (1034,)
    MonoImporter = (1035,)
    AssetServerCache = (1037,)
    LibraryAssetImporter = (1038,)
    ModelImporter = (1040,)
    FBXImporter = (1041,)
    TrueTypeFontImporter = (1042,)
    MovieImporter = (1044,)
    EditorBuildSettings = (1045,)
    DDSImporter = (1046,)
    InspectorExpandedState = (1048,)
    AnnotationManager = (1049,)
    PluginImporter = (1050,)
    EditorUserBuildSettings = (1051,)
    PVRImporter = (1052,)
    ASTCImporter = (1053,)
    KTXImporter = (1054,)
    IHVImageFormatImporter = (1055,)
    AnimatorStateTransition = (1101,)
    AnimatorState = (1102,)
    HumanTemplate = (1105,)
    AnimatorStateMachine = (1107,)
    PreviewAssetType = (1108,)
    AnimatorTransition = (1109,)
    SpeedTreeImporter = (1110,)
    AnimatorTransitionBase = (1111,)
    SubstanceImporter = (1112,)
    LightmapParameters = (1113,)
    LightmapSnapshot = (1120,)
    GISRaster = (1121,)
    GISRasterImporter = (1122,)
    CadImporter = (1123,)
    SketchUpImporter = (1124,)
    BuildReport = (1125,)
    PackedAssets = (1126,)
    VideoClipImporter = (1127,)
    ActivationLogComponent = (2000,)
    # kLargestEditorClassID = 2001
    # kClassIdOutOfHierarchy = 100000
    SubDerived = (367388927,)
    SiblingDerived = (334799969,)
    SpriteAtlas = (687078895,)
    Derived = (1091556383,)
    LowerResBlitTexture = (1480428607,)
    RenderPassAttachment = 1571458007

    def __str__(self):
        return self.name

    def __eq__(self, value):
        if isinstance(value, str):
            return self.name == value
        else:
            return self.value == value

    @classmethod
    def _missing_(cls, value):
        return ClassIDType.UnknownType



from .Object import Object


class PlayerSettings(Object):
    def __init__(self, reader):
        super().__init__(reader=reader)
        if self.version[0] > 5 or (
            self.version[0] == 5 and self.version[1] >= 4
        ):  # 5.4.0 nad up
            self.productGUID = reader.read_bytes(16)

        self.AndroidProfiler = reader.read_boolean()
        # bool AndroidFilterTouchesWhenObscured 2017.2 and up
        # bool AndroidEnableSustainedPerformanceMode 2018 and up
        reader.align_stream()
        self.defaultScreenOrientation = reader.read_int()
        self.targetDevice = reader.read_int()
        if self.version[0] < 5 or (
            self.version[0] == 5 and self.version[1] < 3
        ):  # 5.3 down
            if self.version[0] < 5:  # 5.0 down
                self.targetPlatform = reader.read_int()  # 4.0 and up targetGlesGraphics
                if self.version[0] > 4 or (
                    self.version[0] == 4 and self.version[1] >= 6
                ):  # 4.6 and up
                    self.targetIOSGraphics = reader.read_int()
            self.targetResolution = reader.read_int()
        else:
            self.useOnDemandResources = reader.read_boolean()
            reader.align_stream()
        if self.version[0] > 3 or (
            self.version[0] == 3 and self.version[1] >= 5
        ):  # 3.5 and up
            self.accelerometerFrequency = reader.read_int()
        self.companyName = reader.read_aligned_string()
        self.productName = reader.read_aligned_string()



try:
    from fsb5 import FSB5
except ImportError:
    print("Couldn't import fsb5.\nfsb5 is required to process convert audio clips.")


def extract_audioclip_samples(audio) -> dict:
    """extracts all the sample data from an AudioClip
	Copied from unitypack
	https://github.com/HearthSim/UnityPack/blob/d9ce99fac3c917fa44b0042c9114b7cd03aa9884/unitypack/utils.py#L14


	:param audio: AudioClip
	:type audio: AudioClip
	:return: {filename : sample(bytes)}
	:rtype: dict
	"""
    ret = {}

    if not audio.m_AudioData:
        # eg. StreamedResource not available
        return {}

    af = FSB5(audio.m_AudioData)
    for i, sample in enumerate(af.samples):
        if i > 0:
            filename = "%s-%i.%s" % (audio.name, i, af.get_sample_extension())
        else:
            filename = "%s.%s" % (audio.name, af.get_sample_extension())
        try:
            sample = af.rebuild_sample(sample)
        except ValueError as e:
            print("WARNING: Could not extract %r (%s)" % (audio, e))
            continue
        ret[filename] = sample

    return ret



from .NamedObject import NamedObject
from ..streams import EndianBinaryWriter, EndianBinaryReader


class TextAsset(NamedObject):
    def __init__(self, reader: EndianBinaryReader):
        super().__init__(reader=reader)
        self.script = reader.read_bytes(reader.read_int())

    @property
    def text(self):
        return self.script.decode("utf8")

    @text.setter
    def text(self, val):
        self.script = val.encode("utf8")

    def save(self, writer: EndianBinaryWriter = None):
        if writer is None:
            writer = EndianBinaryWriter(endian=self.reader.endian)
        super().save(writer)
        writer.write_int(len(self.script))
        writer.write_bytes(self.script)
        writer.align_stream()

        self.reader.data = writer.bytes



from enum import IntEnum


class BuildTarget(IntEnum):
    UnknownPlatform = (3716,)
    DashboardWidget = (1,)
    StandaloneOSX = (2,)
    StandaloneOSXPPC = (3,)
    StandaloneOSXIntel = (4,)
    StandaloneWindows = (5,)
    WebPlayer = (6,)
    WebPlayerStreamed = (7,)
    Wii = (8,)
    iOS = (9,)
    PS3 = (10,)
    XBOX360 = (11,)
    Android = (13,)
    StandaloneGLESEmu = (14,)
    NaCl = (16,)
    StandaloneLinux = (17,)
    FlashPlayer = (18,)
    StandaloneWindows64 = (19,)
    WebGL = (20,)
    WSAPlayer = (21,)
    StandaloneLinux64 = (24,)
    StandaloneLinuxUniversal = (25,)
    WP8Player = (26,)
    StandaloneOSXIntel64 = (27,)
    BlackBerry = (28,)
    Tizen = (29,)
    PSP2 = (30,)
    PS4 = (31,)
    PSM = (32,)
    XboxOne = (33,)
    SamsungTV = (34,)
    N3DS = (35,)
    WiiU = (36,)
    tvOS = (37,)
    Switch = (38,)
    NoTarget = -2

    @classmethod
    def _missing_(cls, value):
        return BuildTarget.UnknownPlatform



from . import classes


class ObjectReader:
    def __init__(self, reader, assets_file, object_info):
        self.assets_file = assets_file
        self.path_id = object_info.path_id
        self.byte_start = object_info.byte_start
        self.byte_size = object_info.byte_size
        self.serialized_type = object_info.serialized_type
        self.platform = assets_file.target_platform
        self.version2 = assets_file.header.version
        self.type = object_info.class_id
        self.version = assets_file.version
        self.build_type = assets_file.build_type
        self.reader = reader

    @property
    def container(self):
        return (
            self.assets_file._container[self.path_id]
            if self.path_id in self.assets_file._container
            else None
        )

    def reset(self):
        self.reader.Position = self.byte_start

    def read(self):
        return getattr(classes, self.type.name, classes.Object)(self)

    def __getattr__(self, item: str):
        if hasattr(self.reader, item):
            return getattr(self.reader, item)

    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, self.type.name)



class Vector4:
    X: float
    Y: float
    Z: float
    W: float

    def __init__(self, *args):
        if len(args) == 4:  # float x, float y, float z, float w
            self.X = args[0]
            self.Y = args[1]
            self.Z = args[2]
            self.W = args[3]
        elif len(args) == 2:  # Vector3 value, float w
            self.X = args[0].X
            self.Y = args[0].Y
            self.Z = args[0].Z
            self.W = args[1]

    """
        public float this[int index]
        {
            get
            {
                switch (index)
                {
                    case 0: return X;
                    case 1: return Y;
                    case 2: return Z;
                    case 3: return W;
                    default: throw new ArgumentOutOfRangeException(nameof(index), "Invalid Vector4 index!");
                }
            }

            set
            {
                switch (index)
                {
                    case 0: X = value; break;
                    case 1: Y = value; break;
                    case 2: Z = value; break;
                    case 3: W = value; break;
                    default: throw new ArgumentOutOfRangeException(nameof(index), "Invalid Vector4 index!");
                }
            }
        }


        public override int GetHashCode()
        {
            return X.GetHashCode() ^ (Y.GetHashCode() << 2) ^ (Z.GetHashCode() >> 2) ^ (W.GetHashCode() >> 1);
        }

        public override bool Equals(object other)
        {
            if (!(other is Vector4))
                return false;
            return Equals((Vector4)other);
        }

        public bool Equals(Vector4 other)
        {
            return X.Equals(other.X) && Y.Equals(other.Y) && Z.Equals(other.Z) && W.Equals(other.W);
        }

        public void Normalize()
        {
            var length = Length();
            if (length > kEpsilon)
            {
                var invNorm = 1.0f / length;
                X *= invNorm;
                Y *= invNorm;
                Z *= invNorm;
                W *= invNorm;
            }
            else
            {
                X = 0;
                Y = 0;
                Z = 0;
                W = 0;
            }
        }

        public float Length()
        {
            return (float)Math.Sqrt(LengthSquared());
        }

        public float LengthSquared()
        {
            return X * X + Y * Y + Z * Z + W * W;
        }

        public static Vector4 Zero => new Vector4();

        public static Vector4 operator +(Vector4 a, Vector4 b)
        {
            return new Vector4(a.X + b.X, a.Y + b.Y, a.Z + b.Z, a.W + b.W);
        }

        public static Vector4 operator -(Vector4 a, Vector4 b)
        {
            return new Vector4(a.X - b.X, a.Y - b.Y, a.Z - b.Z, a.W - b.W);
        }

        public static Vector4 operator -(Vector4 a)
        {
            return new Vector4(-a.X, -a.Y, -a.Z, -a.W);
        }

        public static Vector4 operator *(Vector4 a, float d)
        {
            return new Vector4(a.X * d, a.Y * d, a.Z * d, a.W * d);
        }

        public static Vector4 operator *(float d, Vector4 a)
        {
            return new Vector4(a.X * d, a.Y * d, a.Z * d, a.W * d);
        }

        public static Vector4 operator /(Vector4 a, float d)
        {
            return new Vector4(a.X / d, a.Y / d, a.Z / d, a.W / d);
        }

        public static bool operator ==(Vector4 lhs, Vector4 rhs)
        {
            return (lhs - rhs).LengthSquared() < kEpsilon * kEpsilon;
        }

        public static bool operator !=(Vector4 lhs, Vector4 rhs)
        {
            return !(lhs == rhs);
        }

        public static implicit operator Vector2(Vector4 v)
        {
            return new Vector2(v.X, v.Y);
        }

        public static implicit operator Vector3(Vector4 v)
        {
            return new Vector3(v.X, v.Y, v.Z);
        }

        public static implicit operator Color(Vector4 v)
        {
            return new Color(v.X, v.Y, v.Z, v.W);
        }

        private const float kEpsilon = 0.00001F;
    }
}
"""



from .Animation import Animation
from .AnimationClip import AnimationClip
from .Animator import Animator
from .AnimatorController import AnimatorController
from .AnimatorOverrideController import AnimatorOverrideController
from .AssetBundle import AssetBundle
from .AudioClip import AudioClip
from .Avatar import Avatar
from .Behaviour import Behaviour
from .BuildSettings import BuildSettings
from .Component import Component
from .EditorExtension import EditorExtension
from .Font import Font
from .GameObject import GameObject
from .Material import Material
from .Mesh import Mesh
from .MeshFilter import MeshFilter
from .MeshRenderer import MeshRenderer
from .MonoBehaviour import MonoBehaviour
from .MonoScript import MonoScript
from .MovieTexture import MovieTexture
from .NamedObject import NamedObject
from .Object import Object
from .PPtr import PPtr
from .PlayerSettings import PlayerSettings
from .RectTransform import RectTransform
from .Renderer import Renderer
from .RuntimeAnimatorController import RuntimeAnimatorController
from .Shader import Shader
from .SkinnedMeshRenderer import SkinnedMeshRenderer
from .Sprite import Sprite
from .SpriteAtlas import SpriteAtlas
from .TextAsset import TextAsset
from .Texture import Texture
from .Texture2D import Texture2D
from .Transform import Transform
from .VideoClip import VideoClip



import io
import logging
import os
from zipfile import ZipFile

from .EndianBinaryReader import EndianBinaryReader
from .Progress import Progress
from .files import BundleFile, SerializedFile, WebFile
from .helpers import ImportHelper

FileType = ImportHelper.FileType

# create logger
logger = logging.getLogger("UnityPy")
logger.setLevel(logging.DEBUG)

# create console handler and set level to debug
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)

# create formatter
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")

# add formatter to ch
ch.setFormatter(formatter)

# add ch to logger
logger.addHandler(ch)


class AssetsManager:
    assets: dict
    resource_file_readers: dict
    import_files: dict
    Progress: Progress

    def __init__(self, *args):
        self.assets = {}
        self.resource_file_readers = {}
        self.import_files = {}

        self.Progress = Progress()

        if args:
            for arg in args:
                if isinstance(arg, str):
                    if os.path.isfile(arg):
                        if os.path.splitext(arg)[-1] in [".apk", ".zip"]:
                            self.load_zip_file(arg)
                        else:
                            self.load_file(arg)
                    elif os.path.isdir(arg):
                        self.load_folder(arg)
                else:
                    self.load_file(arg)

    def load_files(self, files: list):
        """Loads all files (list) into the AssetsManager and merges .split files for common usage."""
        path = os.path.dirname(files[0])
        ImportHelper.merge_split_assets(path)
        to_read_file = ImportHelper.processing_split_files(files)
        self.load(to_read_file)

    def load_folder(self, path: str):
        """Loads all files in the given path and its subdirs into the AssetsManager."""
        ImportHelper.merge_split_assets(path, True)
        files = ImportHelper.list_all_files(path)
        to_read_file = ImportHelper.processing_split_files(files)
        self.load(to_read_file)

    def load(self, files: list):
        """Loads all files into the AssetsManager."""
        for f in files:
            self.import_files[os.path.basename(f)] = f

        self.Progress.reset()
        # use a for loop because list size can change
        for i, f in enumerate(self.import_files.values()):
            self.load_file(f)
            self.Progress.report(i + 1, len(self.import_files))

    def load_file(self, full_name: str = "", data=None):
        typ, reader = ImportHelper.check_file_type(data if data else full_name)
        if not full_name:
            full_name = str(data)[:256]
        if typ == FileType.AssetsFile:
            self.load_assets_file(full_name, reader)
        elif typ == FileType.BundleFile:
            self.load_bundle_file(full_name, reader)
        elif typ == FileType.WebFile:
            self.load_web_file(full_name, reader)
        elif typ == FileType.ZIP:
            self.load_zip_file(reader.stream)

    def load_zip_file(self, value):
        buffer = None
        if isinstance(value, str) and os.path.exists(value):
            buffer = open(value, "rb")
        elif isinstance(value, (bytes, bytearray)):
            buffer = ZipFile(io.BytesIO(value))
        elif isinstance(value, (io.BufferedReader, io.BufferedIOBase)):
            buffer = value

        z = ZipFile(buffer)

        for path in z.namelist():
            self.load_file(path, z.open(path).read())

    def load_assets_file(self, full_name: str, reader: EndianBinaryReader):
        file_name = os.path.basename(full_name)
        if file_name not in self.assets:
            logging.info(f"Loading {full_name}")
            try:
                assets_file = SerializedFile(self, full_name, reader)
                self.assets[assets_file.name] = assets_file

                for sharedFile in assets_file._externals:
                    shared_file_path = os.path.join(
                        os.path.dirname(full_name), sharedFile.name
                    )
                    shared_file_name = sharedFile.name

                    if shared_file_name not in self.import_files:
                        if not os.path.exists(shared_file_path):
                            find_files = [
                                f
                                for f in ImportHelper.list_all_files(
                                    os.path.dirname(full_name)
                                )
                                if shared_file_name in f
                            ]
                            if find_files:
                                shared_file_path = find_files[0]

                        if os.path.exists(shared_file_path):
                            self.import_files[shared_file_name] = shared_file_path

                return assets_file

            except Exception as e:
                reader.dispose()
                logging.error(f"Unable to load assets file {file_name}", e)
        else:
            reader.dispose()

    def load_assets_from_memory(
        self,
        full_name: str,
        reader: EndianBinaryReader,
        original_path: str,
        unity_version=None,
    ):
        file_name = os.path.basename(full_name)
        if file_name.endswith((".resS", ".resource", ".config", ".xml", ".dat")):
            self.resource_file_readers[file_name] = reader
            return False
        elif file_name not in self.assets:
            try:
                assets_file = SerializedFile(self, full_name, reader)
                assets_file.original_path = original_path
                if assets_file.header.version < 7:
                    assets_file.set_version(unity_version)
                self.assets[file_name] = assets_file
            except Exception as e:
                logging.error(
                    f"Unable to load assets file {file_name} from {original_path}", e
                )
                self.resource_file_readers[file_name] = reader
                return False
        return True

    def load_bundle_file(
        self, full_name: str, reader: EndianBinaryReader, parent_path=None
    ):
        file_name = os.path.basename(full_name)
        logging.info(f"Loading {full_name}")
        bundle_file = None
        try:
            bundle_file = BundleFile(reader, full_name)
            for f in bundle_file.files:
                dummy_path = os.path.join(os.path.dirname(full_name), f.name)
                self.load_assets_from_memory(
                    dummy_path,
                    EndianBinaryReader(f.stream),
                    full_name if parent_path else bundle_file.version_engine,
                )
        except Exception as e:
            string = f"Unable to load bundle file {file_name}"
            if parent_path:
                string += f" from {os.path.basename(parent_path)}"
            string += "\n" + str(e)
            logging.error(string, e)
        finally:
            reader.dispose()
        return bundle_file

    def load_web_file(self, full_name: str, reader: EndianBinaryReader):
        file_name = os.path.basename(full_name)
        logging.info(f"Loading {full_name}")
        dispose = True
        web_file = None
        try:
            web_file = WebFile(reader)
            for f in web_file.files:
                dummy_path = os.path.join(os.path.dirname(full_name), f.name)
                typ, reader = ImportHelper.check_file_type(f.stream)
                if typ == FileType.AssetsFile:
                    dispose = self.load_assets_from_memory(
                        dummy_path, reader, full_name
                    )
                elif typ == FileType.BundleFile:
                    self.load_bundle_file(dummy_path, reader, full_name)
                elif typ == FileType.WebFile:
                    self.load_web_file(dummy_path, reader)
        except Exception as e:
            logging.error(f"Unable to load web file {file_name}", e)
        finally:
            if dispose:
                reader.dispose()
        return web_file

    def clear(self):
        for assetsFile in self.assets:
            assetsFile.Objects = []
            assetsFile.reader.Close()
        self.assets = {}

        for resourceFileReader in self.resource_file_readers:
            resourceFileReader.Value.Close()

        self.resource_file_readers = {}
